from typing import Dict, Tuple
import logging
import pandas as pd
import numpy as np
import warnings
import attr
from attr.validators import instance_of
from enum import Enum
from pysight.nd_hist_generator.movie import ImagingSoftware
from ..line_signal_validators.scanimage import ScanImageLineValidator
from ..line_signal_validators.mscan import MScanLineValidator


class LinesType(Enum):
    SWEEPS_REBUILD = "sweeps-rebuild"
    SWEEPS_FROM_SCRATCH = "sweeps-from-scratch"
    CORRUPT = "corrupt"
    REBUILD = "rebuild"
    SCANIMAGE = "si_valid"
    MSCAN = "mscan_valid"
    FROM_SCRATCH = "from_scratch"


@attr.s(slots=True)
class SignalValidator:
    """
    Parse a dictionary of all recorded signals and verify the data's integrity

    :param dict dict_of_data: Dictionary with keys corresponding to signal type,
                            and values as DataFrames.
    :param list data_to_grab: Column names from the DF of the data to parse.
    :param int num_of_lines: Number of lines (x pixels) in the frame)
    :param int num_of_frames: Number of frames of the data. If "Frames" is not in the
                            dict_of_data keys then this number is disregarded.
    :param float binwidth: Multiscaler binwidth in seconds.
    :param float line_freq: Line frequency of the fast scanner (x) in Hz.
    :param bool bidir: Whether the scan was bidirectional.
    :param float delay_between_frames: Time between subsequent frames. Depends on imaging
                                        software and hardware.
    :param bool use_sweeps: Use the sweeps of the multiscaler as an indicator of new lines.
    :param float bidir_phase: A number in microseconds typically reported by the scanning software.
                                Helps with alignment of the generated image when using a bidirectional scan.
    :param ImagingSoftware imaging_software: Software used for imaging.
    :param float change_thresh: Percent-like [0, 1] threshold to discard irregular line signals.
                                Above which PySight will terminate with an exception.
    :param dict lst_metadata: Acquisiton parameters as recorded in the multiscaler.
    """

    dict_of_data = attr.ib(validator=instance_of(dict))
    data_to_grab = attr.ib(default=["abs_time", "sweep"], validator=instance_of(list))
    num_of_lines = attr.ib(default=512, validator=instance_of(int))
    num_of_frames = attr.ib(default=1, validator=instance_of(int))
    binwidth = attr.ib(default=800e-12, validator=instance_of(float))
    line_freq = attr.ib(default=7900.0, validator=instance_of(float))
    bidir = attr.ib(default=False, validator=instance_of(bool))
    delay_between_frames = attr.ib(default=0.001_135_5, validator=instance_of(float))
    use_sweeps = attr.ib(default=False, validator=instance_of(bool))
    bidir_phase = attr.ib(default=-2.7, validator=instance_of(float))
    image_soft = attr.ib(
        default=ImagingSoftware.SCANIMAGE.value, validator=instance_of(str)
    )
    change_thresh = attr.ib(default=0.3, validator=instance_of(float))
    lst_metadata = attr.ib(factory=dict, validator=instance_of(dict))
    handle_line_cases = attr.ib(init=False)
    last_event_time = attr.ib(init=False)
    line_type_data = attr.ib(init=False)
    line_delta = attr.ib(init=False)

    def __attrs_post_init__(self):
        self.handle_line_cases = {
            LinesType.CORRUPT: self.__corrupt,
            LinesType.FROM_SCRATCH: self.__from_scratch,
            LinesType.SWEEPS_FROM_SCRATCH: self.__sweeps_from_scratch,
            LinesType.REBUILD: self.__rebuild,
            LinesType.SWEEPS_REBUILD: self.__sweeps_rebuild,
            LinesType.SCANIMAGE: self.__si,
            LinesType.MSCAN: self.__mscan,
        }

    @property
    def frame_delay(self):
        return np.uint64(self.delay_between_frames / self.binwidth)

    @property
    def total_sweep_time(self):
        return (
            int(self.lst_metadata["fstchan"])
            + int(self.lst_metadata["data_range"])
            + int(self.lst_metadata["holdafter"])
        )

    @property
    def max_sweep(self):
        return int(self.lst_metadata["swpreset"])

    def run(self):
        """ Main pipeline """
        if "Frames" in self.dict_of_data:
            self.num_of_frames = (
                self.dict_of_data["Frames"].shape[0] + 1
            )  # account for first frame
        self.last_event_time = self.__calc_last_event_time()
        self.line_type_data: LinesType = self.__match_line_data_to_case(
            lines=self.dict_of_data.get("Lines")
        )
        self.dict_of_data, self.line_delta = self.handle_line_cases[
            self.line_type_data
        ]()
        self.dict_of_data = self.__validate_frame_input()
        try:
            self.dict_of_data["Laser"] = self.__validate_laser_input(
                self.dict_of_data["Laser"]
            )
        except KeyError:
            pass

        self.__validate_created_data_channels()

    def __bins_bet_lines(self) -> int:
        """
        Calculate the line difference in bins, according to the frequency
        of the line signal.
        :param line_freq: Frequency of lines in Hz. Regardless of bidirectional scanning.
        :return: int - line delta
        """
        freq_in_bins = 1 / (self.line_freq * self.binwidth)
        return int(freq_in_bins / 2) if self.bidir else freq_in_bins

    def __calc_last_event_time(self) -> int:
        """
        Find the last event time for the experiment. Logic as follows:
        No lines / frames data given: Last event time is the last photon time.
        Only lines data given: The last start-of-frame time is created, and the difference between subsequent frames
        in the data is added.
        Frames data exists: The last frame time plus the difference between subsequent frames is the last event time.
        :return int: Last event time
        """

        if "Frames" in self.dict_of_data:
            last_frame_time = self.dict_of_data["Frames"].loc[:, "abs_time"].iloc[-1]

            if self.dict_of_data["Frames"].shape[0] == 1:
                return int(2 * last_frame_time)
            else:
                frame_diff = int(
                    self.dict_of_data["Frames"].loc[:, "abs_time"].diff().mean()
                )
                return int(last_frame_time + frame_diff)

        if "Lines" in self.dict_of_data:
            num_of_lines_recorded = self.dict_of_data["Lines"].shape[0]
            div, mod = divmod(num_of_lines_recorded, self.num_of_lines)

            # if num_of_lines_recorded > (self.num_of_lines * div):  # excessive number of lines
            #     last_line_of_last_frame = self.dict_of_data['Lines'].loc[:, 'abs_time']\
            #         .iloc[div * self.num_of_lines - 1]
            #     line_diff = int(self.dict_of_data['Lines'].iloc[:, 0].diff().mean())
            #     next_frame_event = self.dict_of_data['Lines'].iloc[div * self.num_of_lines, :]
            #     next_frame_event['abs_time'] += np.uint64(10 * line_diff)
            #     self.dict_of_data['Lines'] = self.dict_of_data['Lines'].loc[
            #                                  self.dict_of_data['Lines'].loc[:, 'abs_time'] <= last_line_of_last_frame,:]
            #     if self.image_soft == ImagingSoftware.SCANIMAGE.value:
            #         self.dict_of_data['Lines'] = self.dict_of_data['Lines'].append(next_frame_event)
            #     return int(last_line_of_last_frame + line_diff)

            if (
                mod == 0
            ):  # number of lines contained exactly in number of lines per frame
                return int(
                    self.dict_of_data["Lines"].loc[:, "abs_time"].iloc[-1]
                    + self.dict_of_data["Lines"].loc[:, "abs_time"].diff().mean()
                )

            elif (num_of_lines_recorded < self.num_of_lines * (div + 1)) and (div > 0):
                last_line_of_last_frame = (
                    self.dict_of_data["Lines"]
                    .loc[:, "abs_time"]
                    .iloc[div * self.num_of_lines - 1]
                )
                line_diff = int(self.dict_of_data["Lines"].iloc[:, 0].diff().mean())
                # missing_lines = self.num_of_lines - mod
                # next_frame_event = self.dict_of_data['Lines'].iloc[div * self.num_of_lines, :]
                # next_frame_event['abs_time'] += np.uint64(10 * line_diff)
                # if self.image_soft == ImagingSoftware.SCANIMAGE.value:
                #     self.dict_of_data['Lines'] = self.dict_of_data['Lines'].append(next_frame_event)
                #
                return int(last_line_of_last_frame + line_diff)

            elif div == 0:
                line_diff = int(self.dict_of_data["Lines"].iloc[:, 0].diff().mean())
                missing_lines = self.num_of_lines - num_of_lines_recorded + 1
                return self.dict_of_data["Lines"].iloc[-1, 0] + (
                    line_diff * missing_lines
                )

        # Just PMT data
        max_pmt1 = self.dict_of_data["PMT1"].loc[:, "abs_time"].max()
        try:
            max_pmt2 = self.dict_of_data["PMT2"].loc[:, "abs_time"].max()
        except KeyError:
            return max_pmt1
        else:
            return max(max_pmt1, max_pmt2)

    def __match_line_data_to_case(self, lines: pd.Series) -> LinesType:
        """
        Find the suitable case for this data regarding the line signal recorded
        :param lines: Line signal
        :return: LinesType enumeration
        """
        if self.use_sweeps and "Lines" in self.dict_of_data.keys():
            return LinesType.SWEEPS_REBUILD

        if self.use_sweeps and "Lines" not in self.dict_of_data.keys():
            return LinesType.SWEEPS_FROM_SCRATCH

        if "Lines" in self.dict_of_data.keys():
            lines = lines.loc[:, "abs_time"].copy()
            # Verify that the input is not corrupt
            if lines.shape[0] < self.num_of_lines // 2:
                logging.warning(
                    "Line data was corrupt - there were too few lines.\n"
                    "Simulating line data using GUI's parameters."
                )

                if "Frames" in self.dict_of_data.keys():
                    return LinesType.REBUILD
                else:
                    return LinesType.CORRUPT

            lines = self.__add_zeroth_line_event(lines=lines)

            # Analyze the line signal
            rel_idx = np.where(
                np.abs(lines.diff().pct_change(periods=1)) > self.change_thresh
            )[0]
            is_corrupt_sig: bool = (len(rel_idx) // 2) / lines.shape[
                0
            ] > self.change_thresh
            if is_corrupt_sig and "Frames" in self.dict_of_data.keys():
                return LinesType.REBUILD
            elif is_corrupt_sig and "Frames" not in self.dict_of_data.keys():
                return LinesType.CORRUPT
            else:  # not corrupt
                if self.image_soft == ImagingSoftware.SCANIMAGE.value:
                    return LinesType.SCANIMAGE
                elif self.image_soft == ImagingSoftware.MSCAN.value:
                    return LinesType.MSCAN

        else:
            return LinesType.FROM_SCRATCH

    def __si(self) -> Tuple[Dict, np.uint64]:
        return ScanImageLineValidator(sig_val=self).run()

    def __mscan(self) -> Tuple[Dict, np.uint64]:
        return MScanLineValidator(sig_val=self).run()

    def __corrupt(self) -> Tuple[Dict, int]:
        """
        Data is corrupted, and no frame channel can help us.
        :return:
        """
        self.line_delta = self.__bins_bet_lines()
        self.dict_of_data["Lines"] = self.__extrapolate_line_data(
            line_point=self.dict_of_data["Lines"].at[0, "abs_time"]
        )
        return self.dict_of_data, self.line_delta

    def __rebuild(self) -> Tuple[Dict, int]:
        """
        Data is corrupted, but we can rebuild lines on top of the frame channel
        :return:
        """
        line_array = self.__create_line_array()
        self.dict_of_data["Lines"] = pd.DataFrame(
            line_array, columns=["abs_time"], dtype="uint64"
        )
        line_delta = self.last_event_time / (
            self.num_of_lines * int(self.num_of_frames)
        )
        return self.dict_of_data, line_delta

    def __add_zeroth_line_event(self, lines: pd.Series) -> pd.Series:
        """
        If lines was used as the starting signal - add a 0 line time
        :param lines:
        :return: Corrected line signal
        """
        line_delta = lines.diff().median()
        zeroth_line_delta = np.abs(lines.iat[0] - line_delta) / line_delta
        if (
            zeroth_line_delta < 0.05
        ):  # first line came exactly line_delta time after the start of the experiment
            lines = pd.Series([0], dtype=np.uint64).append(lines)
            self.dict_of_data["Lines"] = pd.DataFrame(
                [[0] * len(self.data_to_grab)],
                columns=self.data_to_grab,
                dtype=np.uint64,
            ).append(self.dict_of_data["Lines"], ignore_index=True)

        return lines

    def __si_valid(self) -> Tuple[Dict, int]:
        """
        Data is valid. Check whether we need a 0-time line event
        """
        line_delta = self.dict_of_data["Lines"].loc[:, "abs_time"].diff().median()
        # zeroth_line_delta = np.abs(self.dict_of_data['Lines'].loc[0, 'abs_time'] - line_delta)/line_delta
        # if zeroth_line_delta < 0.05:
        #     self.dict_of_data['Lines'] = pd.DataFrame([[0] * len(self.data_to_grab)],
        #                                          columns=self.data_to_grab,
        #                                          dtype='uint64')\
        #         .append(self.dict_of_data['Lines'], ignore_index=True)
        return self.dict_of_data, np.uint64(line_delta)

    def __from_scratch(self) -> Tuple[Dict, int]:
        line_array = self.__create_line_array()
        self.dict_of_data["Lines"] = pd.DataFrame(
            line_array, columns=["abs_time"], dtype="uint64"
        )
        line_delta = self.last_event_time / (
            self.num_of_lines * int(self.num_of_frames)
        )
        return self.dict_of_data, line_delta

    def __sweeps_rebuild(self) -> Tuple[Dict, int]:
        line_delta = 1 / self.line_freq if not self.bidir else 1 / (2 * self.line_freq)
        self.num_of_frames = np.ceil(
            self.last_event_time / (line_delta / self.binwidth * self.num_of_lines)
        )
        line_array = self.__create_line_array(line_delta=line_delta / self.binwidth)
        self.dict_of_data["Lines"] = pd.DataFrame(
            line_array, columns=["abs_time"], dtype="uint64"
        )
        return self.dict_of_data, int(line_delta / self.binwidth)

    def __sweeps_from_scratch(self) -> Tuple[Dict, int]:
        sweep_vec = np.arange(self.max_sweep + 1, dtype=np.uint64)
        if len(sweep_vec) < 2:
            logging.warning(
                "All data was registered to a single sweep. Line data will be completely simulated."
            )
        else:
            self.dict_of_data["Lines"] = pd.DataFrame(
                sweep_vec * self.total_sweep_time, columns=["abs_time"], dtype=np.uint64
            )
        return self.dict_of_data, self.total_sweep_time

    def __extrapolate_line_data(self, line_point: int = 0) -> pd.DataFrame:
        """
        From a single line signal extrapolate the presumed line data vector. The line frequency is double
        the original frequency. If needed, events will be discarded later.
        :param line_point: Start interpolation from this point.
        :return: pd.DataFrame of line data
        """
        # Verify input
        if (
            line_point / self.line_delta > self.num_of_lines
        ):  # The first recorded line came very late
            line_point = 0

        # Create the matrix containing the duplicate frame data
        delay_between_frames_in_bins = int(self.delay_between_frames / self.binwidth)
        time_of_frame = (
            self.line_delta * self.num_of_lines + delay_between_frames_in_bins
        )
        num_of_frames = max(int(np.floor(self.last_event_time / time_of_frame)), 1)
        time_of_frame_mat = np.tile(
            time_of_frame * np.arange(num_of_frames), (self.num_of_lines, 1)
        )

        # Create the matrix containing the duplicate line data
        line_vec = np.arange(
            start=line_point,
            stop=self.line_delta * self.num_of_lines,
            step=self.line_delta,
            dtype=np.uint64,
        )
        line_vec = np.r_[
            np.flip(
                np.arange(
                    start=line_point, stop=0, step=-self.line_delta, dtype=np.uint64
                )[1:],
                axis=0,
            ),
            line_vec,
        ]
        # Check if 0 should be included
        if line_vec[0] - self.line_delta == 0:
            line_vec = np.r_[0, line_vec]

        line_mat = np.tile(line_vec.reshape(len(line_vec), 1), (1, num_of_frames))

        # Add them up
        assert len(line_mat) > 0
        assert len(time_of_frame_mat) > 0
        final_mat = line_mat + time_of_frame_mat
        line_vec_final = np.ravel(final_mat, order="F")

        return pd.DataFrame(line_vec_final, columns=["abs_time"], dtype=np.uint64)

    def __create_line_array(self, line_delta: int = 0) -> np.ndarray:
        """Create a pandas Series of start-of-line times"""

        total_lines = self.num_of_lines * int(self.num_of_frames)
        if line_delta == 0:
            line_array = np.arange(
                start=0,
                stop=self.last_event_time,
                step=self.last_event_time / total_lines,
                dtype=np.uint64,
            )
        else:
            line_array = np.arange(
                start=0, stop=self.last_event_time, step=line_delta, dtype=np.uint64
            )
        return line_array

    def __validate_frame_input(self):

        if "Frames" in self.dict_of_data.keys():
            if self.image_soft == "MScan":
                self.dict_of_data["Frames"] = pd.DataFrame(
                    [[0] * len(self.data_to_grab)],
                    columns=self.data_to_grab,
                    dtype="uint64",
                ).append(self.dict_of_data["Frames"], ignore_index=True)
        else:
            frame_array = self.__create_frame_array(
                lines=self.dict_of_data["Lines"].loc[:, "abs_time"]
            )
            self.dict_of_data["Frames"] = pd.DataFrame(
                frame_array, columns=["abs_time"], dtype="uint64"
            )

        return self.dict_of_data

    def __create_frame_array(self, lines: pd.Series = None) -> np.ndarray:
        """Create a pandas Series of start-of-frame times"""

        num_of_recorded_lines = lines.shape[0]
        actual_num_of_frames = max(num_of_recorded_lines // self.num_of_lines, 1)

        if num_of_recorded_lines < self.num_of_lines:
            array_of_frames = np.linspace(
                start=0,
                stop=self.last_event_time,
                num=int(actual_num_of_frames),
                endpoint=False,
                dtype=np.uint64,
            )
        else:
            unnecess_lines = num_of_recorded_lines % self.num_of_lines
            array_of_frames = lines.iloc[
                0 : int(num_of_recorded_lines - unnecess_lines) : self.num_of_lines
            ]

        return np.array(array_of_frames)

    def __validate_created_data_channels(self) -> None:
        """
        Make sure that the dictionary that contains all data channels makes sense.
        """
        assert {"PMT1", "Lines", "Frames"} <= set(
            self.dict_of_data.keys()
        )  # A is subset of B

        if (
            self.dict_of_data["Frames"].shape[0] > self.dict_of_data["Lines"].shape[0]
        ):  # more frames than lines
            raise UserWarning("More frames than lines, consider replacing the two.")

        try:
            if (
                self.dict_of_data["TAG Lens"].shape[0]
                < self.dict_of_data["Lines"].shape[0]
            ):
                raise UserWarning(
                    "More lines than TAG pulses, consider replacing the two."
                )
        except KeyError:
            pass

        try:
            if (
                self.dict_of_data["Laser"].shape[0]
                < self.dict_of_data["Lines"].shape[0]
                or self.dict_of_data["Laser"].shape[0]
                < self.dict_of_data["Frames"].shape[0]
            ):
                raise UserWarning(
                    "Laser pulses channel contained less ticks than the Lines or Frames channel."
                )
        except KeyError:
            pass

        try:
            if (
                self.dict_of_data["Laser"].shape[0]
                < self.dict_of_data["TAG Lens"].shape[0]
            ):
                raise UserWarning(
                    "Laser pulses channel contained less ticks than the TAG lens channel."
                )
        except KeyError:
            pass

    def __validate_laser_input(self, pulses) -> pd.Series:
        """
        Create an orderly laser pulse train.
        :param pulses: Laser data
        :return:
        """

        diffs = pulses.loc[:, "abs_time"].diff()
        rel_idx = (diffs <= 1.05 * np.ceil((1 / (self.laser_freq * self.binwidth)))) & (
            diffs >= 0.95 * np.floor((1 / (self.laser_freq * self.binwidth)))
        )
        pulses_final = pulses[rel_idx]  # REMINDER: Laser offset wasn't added
        if len(pulses_final) < 0.9 * len(
            pulses
        ):  # TODO: If there's holdafter time, there's a good chance that the pulses
            # will not be periodic due to this extra time.
            logging.warning(
                "More than 10% of pulses were filtered due to bad timings. Make sure the laser input is fine."
            )

        pulses_final = pd.concat(
            [pulses.loc[:0, :], pulses_final]
        )  # Add back the first pulse
        pulses_final.reset_index(drop=True, inplace=True)

        return pulses_final

    def add_phase_to_bidir_lines(self, lines: pd.Series):
        """
        "Fix" temporarily the lines for them to pass the corruption check
        :param lines: Lines after the phase change
        :param bidir_phase:
        :param binwidth:
        :return: Lines "pre-fix" of phase change - their original form
        """
        phase_in_seconds = self.bidir_phase * 1e-6
        if phase_in_seconds < 0:
            lines.iloc[1::2] -= np.uint64(np.abs(phase_in_seconds / self.binwidth))
        else:
            lines.iloc[1::2] += np.uint64(phase_in_seconds / self.binwidth)
        return lines



from collections import OrderedDict
import numpy as np
from typing import List
import attr


@attr.s
class Struct(object):
    """ Basic struct-like object for data keeping. """

    start = attr.ib()
    end = attr.ib()
    num = attr.ib(default=None)


class ChoiceManagerHex:
    """
    Switch-like helper class. Contains the type of information
    held in each bit of a single event from the multiscaler, for all timepatches.
    """

    def __init__(self):
        self.__choice_table = {
            "0": self.timepatch_0,
            "5": self.timepatch_5,
            "1": self.timepatch_1,
            "1a": self.timepatch_1a,
            "2a": self.timepatch_2a,
            "22": self.timepatch_22,
            "32": self.timepatch_32,
            "2": self.timepatch_2,
            "5b": self.timepatch_5b,
            "Db": self.timepatch_Db,
            "f3": self.timepatch_f3,
            "43": self.timepatch_43,
            "c3": self.timepatch_c3,
            "3": self.timepatch_3,
        }

    def timepatch_0(self, dict_of_slices):
        dict_of_slices.pop("tag", None)
        dict_of_slices.pop("sweep", None)

        dict_of_slices["time_rel_sweep"].start = 0
        dict_of_slices["time_rel_sweep"].end = 3
        dict_of_slices["time_rel_sweep"].needs_bits = False

        dict_of_slices["chan_edge"].start = 3
        dict_of_slices["chan_edge"].end = 4

        return dict_of_slices

    def timepatch_5(self, dict_of_slices):
        dict_of_slices.pop("tag", None)

        dict_of_slices["sweep"].start = 0
        dict_of_slices["sweep"].end = 2
        dict_of_slices["sweep"].needs_bits = False

        dict_of_slices["time_rel_sweep"].start = 2
        dict_of_slices["time_rel_sweep"].end = 7
        dict_of_slices["time_rel_sweep"].needs_bits = False

        dict_of_slices["chan_edge"].start = 7
        dict_of_slices["chan_edge"].end = 8

        return dict_of_slices

    def timepatch_1(self, dict_of_slices):
        dict_of_slices.pop("tag", None)
        dict_of_slices.pop("sweep", None)

        dict_of_slices["time_rel_sweep"].start = 0
        dict_of_slices["time_rel_sweep"].end = 7
        dict_of_slices["time_rel_sweep"].needs_bits = False

        dict_of_slices["chan_edge"].start = 7
        dict_of_slices["chan_edge"].end = 8

        return dict_of_slices

    def timepatch_1a(self, dict_of_slices):
        dict_of_slices.pop("tag", None)

        dict_of_slices["sweep"].start = 0
        dict_of_slices["sweep"].end = 4
        dict_of_slices["sweep"].needs_bits = False

        dict_of_slices["time_rel_sweep"].start = 4
        dict_of_slices["time_rel_sweep"].end = 11
        dict_of_slices["time_rel_sweep"].needs_bits = False

        dict_of_slices["chan_edge"].start = 11
        dict_of_slices["chan_edge"].end = 12

        return dict_of_slices

    def timepatch_2a(self, dict_of_slices):
        dict_of_slices["tag"].start = 0
        dict_of_slices["tag"].end = 2
        dict_of_slices["tag"].needs_bits = False

        dict_of_slices["sweep"].start = 2
        dict_of_slices["sweep"].end = 4
        dict_of_slices["sweep"].needs_bits = False

        dict_of_slices["time_rel_sweep"].start = 4
        dict_of_slices["time_rel_sweep"].end = 11
        dict_of_slices["time_rel_sweep"].needs_bits = False

        dict_of_slices["chan_edge"].start = 11
        dict_of_slices["chan_edge"].end = 12

        return dict_of_slices

    def timepatch_22(self, dict_of_slices):
        dict_of_slices.pop("sweep", None)

        dict_of_slices["tag"].start = 0
        dict_of_slices["tag"].end = 2
        dict_of_slices["tag"].needs_bits = False

        dict_of_slices["time_rel_sweep"].start = 2
        dict_of_slices["time_rel_sweep"].end = 11
        dict_of_slices["time_rel_sweep"].needs_bits = False

        dict_of_slices["chan_edge"].start = 11
        dict_of_slices["chan_edge"].end = 12

        return dict_of_slices

    def timepatch_32(self, dict_of_slices):
        dict_of_slices["lost"] = True
        dict_of_slices.pop("tag", None)

        dict_of_slices["sweep"].start = 0
        dict_of_slices["sweep"].end = 2
        dict_of_slices["sweep"].needs_bits = True

        dict_of_slices["time_rel_sweep"].start = 2
        dict_of_slices["time_rel_sweep"].end = 11
        dict_of_slices["time_rel_sweep"].needs_bits = False

        dict_of_slices["chan_edge"].start = 11
        dict_of_slices["chan_edge"].end = 12

        return dict_of_slices

    def timepatch_2(self, dict_of_slices):
        dict_of_slices.pop("tag", None)
        dict_of_slices.pop("sweep", None)

        dict_of_slices["time_rel_sweep"].start = 0
        dict_of_slices["time_rel_sweep"].end = 11
        dict_of_slices["time_rel_sweep"].needs_bits = False

        dict_of_slices["chan_edge"].start = 11
        dict_of_slices["chan_edge"].end = 12

        return dict_of_slices

    def timepatch_5b(self, dict_of_slices):
        dict_of_slices["lost"] = True

        dict_of_slices["tag"].start = 0
        dict_of_slices["tag"].end = 4
        dict_of_slices["tag"].needs_bits = True

        dict_of_slices["sweep"].start = 4
        dict_of_slices["sweep"].end = 8
        dict_of_slices["sweep"].needs_bits = False

        dict_of_slices["time_rel_sweep"].start = 8
        dict_of_slices["time_rel_sweep"].end = 15
        dict_of_slices["time_rel_sweep"].needs_bits = False

        dict_of_slices["chan_edge"].start = 15
        dict_of_slices["chan_edge"].end = 16

        return dict_of_slices

    def timepatch_Db(self, dict_of_slices):
        dict_of_slices["tag"].start = 0
        dict_of_slices["tag"].end = 4
        dict_of_slices["tag"].needs_bits = False

        dict_of_slices["sweep"].start = 4
        dict_of_slices["sweep"].end = 8
        dict_of_slices["sweep"].needs_bits = False

        dict_of_slices["time_rel_sweep"].start = 8
        dict_of_slices["time_rel_sweep"].end = 15
        dict_of_slices["time_rel_sweep"].needs_bits = False

        dict_of_slices["chan_edge"].start = 15
        dict_of_slices["chan_edge"].end = 16

        return dict_of_slices

    def timepatch_f3(self, dict_of_slices):
        dict_of_slices["lost"] = True

        dict_of_slices["tag"].start = 0
        dict_of_slices["tag"].end = 4
        dict_of_slices["tag"].needs_bits = False

        dict_of_slices["sweep"].start = 4
        dict_of_slices["sweep"].end = 6
        dict_of_slices["sweep"].needs_bits = True

        dict_of_slices["time_rel_sweep"].start = 6
        dict_of_slices["time_rel_sweep"].end = 15
        dict_of_slices["time_rel_sweep"].needs_bits = False

        dict_of_slices["chan_edge"].start = 15
        dict_of_slices["chan_edge"].end = 16

        return dict_of_slices

    def timepatch_43(self, dict_of_slices):
        dict_of_slices.pop("sweep", None)

        dict_of_slices["lost"] = True

        dict_of_slices["tag"].start = 0
        dict_of_slices["tag"].end = 4
        dict_of_slices["tag"].needs_bits = True

        dict_of_slices["time_rel_sweep"].start = 4
        dict_of_slices["time_rel_sweep"].end = 15
        dict_of_slices["time_rel_sweep"].needs_bits = False

        dict_of_slices["chan_edge"].start = 15
        dict_of_slices["chan_edge"].end = 16

        return dict_of_slices

    def timepatch_c3(self, dict_of_slices):
        dict_of_slices.pop("sweep", None)

        dict_of_slices["tag"].start = 0
        dict_of_slices["tag"].end = 4
        dict_of_slices["tag"].needs_bits = False

        dict_of_slices["time_rel_sweep"].start = 4
        dict_of_slices["time_rel_sweep"].end = 15
        dict_of_slices["time_rel_sweep"].needs_bits = False

        dict_of_slices["chan_edge"].start = 15
        dict_of_slices["chan_edge"].end = 16

        return dict_of_slices

    def timepatch_3(self, dict_of_slices):
        # NOT IMPLEMENTED IN THE ACTUAL CODE
        dict_of_slices.pop("sweep", None)

        dict_of_slices["tag"].start = 0
        dict_of_slices["tag"].end = 2
        dict_of_slices["tag"].needs_bits = True

        dict_of_slices["time_rel_sweep"].start = 2
        dict_of_slices["time_rel_sweep"].end = 15
        dict_of_slices["time_rel_sweep"].needs_bits = False

        dict_of_slices["chan_edge"].start = 15
        dict_of_slices["chan_edge"].end = 16

        return dict_of_slices

    def __generate_dict_from_indices(self) -> OrderedDict:
        """
        Generate a basic template dictionary to be populated with slicing indices.
        :param slices:
        :return:
        """
        slice_dict: OrderedDict = OrderedDict()
        keys = ["lost", "tag", "sweep", "time_rel_sweep", "chan_edge"]

        slice_dict[keys[0]] = False
        for key in keys[1:]:
            slice_dict[key] = Struct(np.nan, np.nan)

        return slice_dict

    def process(self, case):
        """
        Generate a dictionary and populate it with the proper value, accoring to the timepatch (==case).
        """
        assert isinstance(case, str)

        dict_of_slices = self.__generate_dict_from_indices()
        populated_dict_of_slices = self.__choice_table[case](dict_of_slices)

        return populated_dict_of_slices


class ChoiceManagerBinary:
    def __init__(self) -> None:
        self.__choice_table = {
            "0": self.timepatch_0,
            "5": self.timepatch_5,
            "1": self.timepatch_1,
            "1a": self.timepatch_1a,
            "2a": self.timepatch_2a,
            "22": self.timepatch_22,
            "32": self.timepatch_32,
            "2": self.timepatch_2,
            "5b": self.timepatch_5b,
            "Db": self.timepatch_Db,
            "f3": self.timepatch_f3,
            "43": self.timepatch_43,
            "c3": self.timepatch_c3,
            "3": self.timepatch_3,
        }

    def timepatch_0(self, dict_of_slices) -> OrderedDict:
        dict_of_slices.pop("tag", None)
        dict_of_slices.pop("sweep", None)

        dict_of_slices["time_rel_sweep"].start = 0
        dict_of_slices["time_rel_sweep"].end = 12
        dict_of_slices["time_rel_sweep"].cols = 16

        dict_of_slices["edge"].start = 12
        dict_of_slices["edge"].end = 13
        dict_of_slices["edge"].cols = 8

        dict_of_slices["chan"].start = 13
        dict_of_slices["chan"].end = 15
        dict_of_slices["chan"].cols = 8

        return dict_of_slices

    def timepatch_5(self, dict_of_slices) -> OrderedDict:
        dict_of_slices.pop("tag", None)

        dict_of_slices["sweep"].start = 0
        dict_of_slices["sweep"].end = 8

        dict_of_slices["time_rel_sweep"].start = 8
        dict_of_slices["time_rel_sweep"].end = 28

        dict_of_slices["chan_edge"].start = 28
        dict_of_slices["chan_edge"].end = 32

        return dict_of_slices

    def timepatch_1(self, dict_of_slices) -> OrderedDict:
        dict_of_slices.pop("tag", None)
        dict_of_slices.pop("sweep", None)

        dict_of_slices["time_rel_sweep"].start = 0
        dict_of_slices["time_rel_sweep"].end = 28

        dict_of_slices["chan_edge"].start = 28
        dict_of_slices["chan_edge"].end = 32

        return dict_of_slices

    def timepatch_1a(self, dict_of_slices) -> OrderedDict:
        dict_of_slices.pop("tag", None)

        dict_of_slices["sweep"].start = 0
        dict_of_slices["sweep"].end = 16

        dict_of_slices["time_rel_sweep"].start = 16
        dict_of_slices["time_rel_sweep"].end = 44

        dict_of_slices["chan_edge"].start = 44
        dict_of_slices["chan_edge"].end = 48

        return dict_of_slices

    def timepatch_2a(self, dict_of_slices) -> OrderedDict:
        dict_of_slices["tag"].start = 0
        dict_of_slices["tag"].end = 8

        dict_of_slices["sweep"].start = 8
        dict_of_slices["sweep"].end = 16

        dict_of_slices["time_rel_sweep"].start = 16
        dict_of_slices["time_rel_sweep"].end = 44

        dict_of_slices["chan_edge"].start = 44
        dict_of_slices["chan_edge"].end = 48

        return dict_of_slices

    def timepatch_22(self, dict_of_slices) -> OrderedDict:
        dict_of_slices.pop("sweep", None)

        dict_of_slices["tag"].start = 0
        dict_of_slices["tag"].end = 8

        dict_of_slices["time_rel_sweep"].start = 8
        dict_of_slices["time_rel_sweep"].end = 44

        dict_of_slices["chan_edge"].start = 44
        dict_of_slices["chan_edge"].end = 48

        return dict_of_slices

    def timepatch_32(self, dict_of_slices) -> OrderedDict:
        dict_of_slices["lost"] = 1
        dict_of_slices.pop("tag", None)

        dict_of_slices["sweep"].start = 1
        dict_of_slices["sweep"].end = 8

        dict_of_slices["time_rel_sweep"].start = 8
        dict_of_slices["time_rel_sweep"].end = 44

        dict_of_slices["chan_edge"].start = 44
        dict_of_slices["chan_edge"].end = 48

        return dict_of_slices

    def timepatch_2(self, dict_of_slices) -> OrderedDict:
        dict_of_slices.pop("tag", None)
        dict_of_slices.pop("sweep", None)

        dict_of_slices["time_rel_sweep"].start = 0
        dict_of_slices["time_rel_sweep"].end = 44

        dict_of_slices["chan_edge"].start = 44
        dict_of_slices["chan_edge"].end = 48

        return dict_of_slices

    def timepatch_5b(self, dict_of_slices) -> OrderedDict:
        dict_of_slices["lost"].start = 0
        dict_of_slices["lost"].end = 1
        dict_of_slices["lost"].cols = 8

        dict_of_slices["tag"].start = 1
        dict_of_slices["tag"].end = 16
        dict_of_slices["tag"].cols = 16

        dict_of_slices["sweep"].start = 16
        dict_of_slices["sweep"].end = 32

        dict_of_slices["time_rel_sweep"].start = 32
        dict_of_slices["time_rel_sweep"].end = 60
        dict_of_slices["time_rel_sweep"].cols = 32

        dict_of_slices["edge"].start = 60
        dict_of_slices["edge"].end = 61
        dict_of_slices["edge"].cols = 8

        dict_of_slices["chan"].start = 61
        dict_of_slices["chan"].end = 64
        dict_of_slices["chan"].cols = 8

        return dict_of_slices

    def timepatch_Db(self, dict_of_slices) -> OrderedDict:
        dict_of_slices["tag"].start = 0
        dict_of_slices["tag"].end = 16

        dict_of_slices["sweep"].start = 16
        dict_of_slices["sweep"].end = 32

        dict_of_slices["time_rel_sweep"].start = 32
        dict_of_slices["time_rel_sweep"].end = 60

        dict_of_slices["chan_edge"].start = 60
        dict_of_slices["chan_edge"].end = 64

        return dict_of_slices

    def timepatch_f3(self, dict_of_slices) -> OrderedDict:
        dict_of_slices["lost"].start = 0
        dict_of_slices["lost"].end = 1
        dict_of_slices["lost"].cols = 8

        dict_of_slices["tag"].start = 1
        dict_of_slices["tag"].end = 17
        dict_of_slices["tag"].cols = 16

        dict_of_slices["sweep"].start = 17
        dict_of_slices["sweep"].end = 24
        dict_of_slices["sweep"].cols = 8

        dict_of_slices["time_rel_sweep"].start = 24
        dict_of_slices["time_rel_sweep"].end = 60
        dict_of_slices["time_rel_sweep"].cols = 40

        dict_of_slices["edge"].start = 60
        dict_of_slices["edge"].end = 61
        dict_of_slices["edge"].cols = 8

        dict_of_slices["chan"].start = 61
        dict_of_slices["chan"].end = 64
        dict_of_slices["chan"].cols = 8

        return dict_of_slices

    def timepatch_43(self, dict_of_slices) -> OrderedDict:
        dict_of_slices.pop("sweep", None)

        dict_of_slices["lost"] = 1

        dict_of_slices["tag"].start = 1
        dict_of_slices["tag"].end = 16

        dict_of_slices["time_rel_sweep"].start = 16
        dict_of_slices["time_rel_sweep"].end = 60

        dict_of_slices["chan_edge"].start = 60
        dict_of_slices["chan_edge"].end = 64

        return dict_of_slices

    def timepatch_c3(self, dict_of_slices) -> OrderedDict:
        dict_of_slices.pop("sweep", None)

        dict_of_slices["tag"].start = 0
        dict_of_slices["tag"].end = 16

        dict_of_slices["time_rel_sweep"].start = 16
        dict_of_slices["time_rel_sweep"].end = 60

        dict_of_slices["chan_edge"].start = 60
        dict_of_slices["chan_edge"].end = 64

        return dict_of_slices

    def timepatch_3(self, dict_of_slices) -> OrderedDict:
        dict_of_slices.pop("sweep", None)

        dict_of_slices["lost"] = 1

        dict_of_slices["tag"].start = 1
        dict_of_slices["tag"].end = 6

        dict_of_slices["time_rel_sweep"].start = 6
        dict_of_slices["time_rel_sweep"].end = 60

        dict_of_slices["chan_edge"].start = 60
        dict_of_slices["chan_edge"].end = 64

        return dict_of_slices

    def __generate_dict_from_indices(self) -> OrderedDict:
        """
        Generate a basic template dictionary to be populated with slicing indices.
        :param slices:
        :return:
        """
        from pysight.nd_hist_generator.movie import Struct

        slice_dict: OrderedDict = OrderedDict()
        keys: List[str] = ["lost", "tag", "sweep", "time_rel_sweep", "edge", "chan"]

        for key in keys:
            slice_dict[key] = Struct(np.nan, np.nan)

        return slice_dict

    def process(self, case: str) -> OrderedDict:
        """
        Generate a dictionary and populate it with the proper value, accoring to the timepatch (==case).
        """
        assert isinstance(case, str)

        dict_of_slices: OrderedDict = self.__generate_dict_from_indices()
        populated_dict_of_slices: OrderedDict = self.__choice_table[case](
            dict_of_slices
        )

        return populated_dict_of_slices



"""
__author__ = Hagai Hargil
"""

from unittest import TestCase
from pysight.nd_hist_generator.outputs import *
from os import sep

import pytest


class TestOutput(TestCase):
    """ Test how well we parse the outputs """

    parser_objects = []
    parser_objects.append(
        OutputParser(output_dict={}, filename="tests_data" + sep + "output_1.hdf5",)
    )
    parser_objects.append(
        OutputParser(
            output_dict={"memory": True},
            filename="tests_data" + sep + "output_2.hdf5",
            reprate=160.3e6,
            z_pixels=10,
            num_of_frames=2,
        )
    )
    parser_objects.append(
        OutputParser(
            output_dict={},
            filename="tests_data" + sep + "output_1.hdf5",
            num_of_frames=2,
        )
    )

    def test_determine_shape(self):
        shapes = [(1, 512, 512), (2, 512, 512, 10), (2, 512, 512)]
        for shape, obj in zip(shapes, self.parser_objects):
            self.assertEqual(shape, obj.determine_data_shape_full())

    def test_non_squeezed_shapes(self):
        shapes = [
            (1, 512, 512, 16, 1),
            (1, 1, 512, 100),
            (10, 512, 16),
            (10, 512, 1),
            (1, 512, 512, 100),
        ]
        squeezed_shapes = [
            (1, 512, 512, 16),
            (1, 512, 100),
            (10, 512, 16),
            (10, 512),
            (1, 512, 512, 100),
        ]
        output_obj = []
        output_obj.append(
            OutputParser(
                output_dict={},
                filename="tests_data" + sep + f"output_3.hdf5",
                z_pixels=16,
                num_of_frames=1,
            )
        )
        output_obj.append(
            OutputParser(
                output_dict={},
                filename="tests_data" + sep + f"output_4.hdf5",
                z_pixels=100,
                num_of_frames=1,
                x_pixels=1,
            )
        )
        output_obj.append(
            OutputParser(
                output_dict={},
                filename="tests_data" + sep + f"output_5.hdf5",
                z_pixels=16,
                num_of_frames=10,
                y_pixels=1,
            )
        )
        output_obj.append(
            OutputParser(
                output_dict={},
                filename="tests_data" + sep + f"output_6.hdf5",
                num_of_frames=10,
                y_pixels=1,
            )
        )
        output_obj.append(
            OutputParser(
                output_dict={},
                filename="tests_data" + sep + f"output_7.hdf5",
                num_of_frames=1,
                z_pixels=100,
            )
        )
        for shape, obj in zip(squeezed_shapes, output_obj):
            self.assertEqual(shape, obj.determine_data_shape_full())


@pytest.fixture
def generate_output_obj():
    def _generate(shape, flim, di=None):
        if di is None:
            di = {1: np.array([1, 2, 3])}
        photons = pd.DataFrame({1: np.array([1, 2, 3])})
        summed_mem = di
        stack = di
        channels = pd.CategoricalIndex([1])
        return PySightOutput(photons, summed_mem, stack, channels, shape, False, dict())

    return _generate


class TestPySightOutput:
    def test_data_shape_txy(self, generate_output_obj):
        shape = (10, 10, 10)
        flim = False
        obj = generate_output_obj(shape, flim)
        shape = DataShape(10, 10, 10, None)
        assert shape == obj._parse_data_shape()

    def test_data_shape_txyz(self, generate_output_obj):
        shape = (10, 10, 10, 10)
        flim = False
        di = {1: np.random.random(shape)}
        obj = generate_output_obj(shape, flim, di)
        shape = DataShape(10, 10, 10, 10)
        assert shape == obj._parse_data_shape()

    def test_data_shape_txyztau(self, generate_output_obj):
        shape = (10, 10, 10, 10)
        flim = True
        di = {1: np.random.random(shape)}
        obj = generate_output_obj(shape, flim, di)
        shape = DataShape(10, 10, 10, 10)
        assert shape == obj._parse_data_shape()

    @pytest.mark.xfail
    def test_data_shape_txyz_notau(self, generate_output_obj):
        shape = (10, 10, 10, 10, 5)
        flim = False
        di = {1: np.random.random(shape)}
        obj = generate_output_obj(shape, flim, di)
        shape = DataShape(10, 10, 10, 10)
        assert shape == obj._parse_data_shape()

    def test_data_shape_txy_tau(self, generate_output_obj):
        shape = (10, 10, 10, 5)
        flim = True
        di = {1: np.random.random(shape)}
        obj = generate_output_obj(shape, flim, di)
        shape = DataShape(10, 10, 10, 5)
        assert shape == obj._parse_data_shape()



import numpy as np

from pysight.ascii_list_file_parser.tabulation import Tabulate, slice_string_arrays


class TestTabulation:

    tab = Tabulate(dict_of_inputs={}, data=np.array([]), dict_of_slices_hex={})

    def test_conversion_hex_to_bits(self):
        diction = {
            "0": "0000",
            "1": "0001",
            "2": "0010",
            "3": "0011",
            "4": "0100",
            "5": "0101",
            "6": "0110",
            "7": "0111",
            "8": "1000",
            "9": "1001",
            "a": "1010",
            "b": "1011",
            "c": "1100",
            "d": "1101",
            "e": "1110",
            "f": "1111",
        }
        assert diction == self.tab.hex_to_bin_dict()

    def test_slice_string_array_1(self):
        a = np.array(["hello", "how", "are", "you"])
        assert np.array_equal(
            slice_string_arrays(a, 1, 3),
            np.array(["el", "ow", "re", "ou"], dtype="|U2"),
        )

    def test_slice_string_array_2(self):
        a = np.array(["hello", "how", "are", "you"])
        assert np.array_equal(
            slice_string_arrays(a, 0, 3),
            np.array(["hel", "how", "are", "you"], dtype="|U3"),
        )



from typing import Dict, Any
import pathlib

import toml
import attr
from attr.validators import instance_of


@attr.s
class Config:
    """
    Load, parse and save a TOML config file.
    """

    config_data = attr.ib(validator=instance_of(dict))

    @classmethod
    def from_gui(cls, gui):
        config_data = {
            "cfg_title": gui.save_as.get(),
            "inputs": {
                "stop1": gui.input_stop1.get(),
                "stop2": gui.input_stop2.get(),
                "stop3": gui.input_stop3.get(),
                "stop4": gui.input_stop4.get(),
                "stop5": gui.input_stop5.get(),
                "start": gui.input_start.get(),
            },
            "image": {
                "num_of_frames": gui.num_of_frames.get(),
                "x_pixels": gui.x_pixels.get(),
                "y_pixels": gui.y_pixels.get(),
                "z_pixels": gui.z_pixels.get(),
                "imaging_software": gui.imaging_software.get(),
            },
            "outputs": {
                "data_filename": gui.filename.get(),
                "summed": gui.summed.get(),
                "memory": gui.memory.get(),
                "stack": gui.stack.get(),
                "flim": gui.flim.get(),
            },
            "advanced": {
                "debug": gui.debug.get(),
                "phase": gui.phase.get(),
                "reprate": gui.reprate.get(),
                "gating": gui.gating.get(),
                "binwidth": gui.binwidth.get(),
                "tag_freq": gui.tag_freq.get(),
                "tag_pulses": gui.tag_pulses.get(),
                "tag_offset": gui.tag_offset.get(),
                "fill_frac": gui.fill_frac.get(),
                "bidir": gui.bidir.get(),
                "keep_unidir": gui.keep_unidir.get(),
                "flim": gui.flim.get(),
                "censor": gui.censor.get(),
                "line_freq": gui.line_freq.get(),
                "sweeps_as_lines": gui.sweeps_as_lines.get(),
                "frame_delay": gui.frame_delay.get(),
                "interleaved": gui.interleaved.get(),
            },
            "tagbits": {
                "tag_bits": gui.tag_bits.get(),
                "group1": {
                    "label1": gui.bits_grp_1_label.get(),
                    "start1": gui.bits_grp_1_start.get(),
                    "end1": gui.bits_grp_1_end.get(),
                },
                "group2": {
                    "label2": gui.bits_grp_2_label.get(),
                    "start2": gui.bits_grp_2_start.get(),
                    "end2": gui.bits_grp_2_end.get(),
                },
                "group3": {
                    "label3": gui.bits_grp_3_label.get(),
                    "start3": gui.bits_grp_3_start.get(),
                    "end3": gui.bits_grp_3_end.get(),
                },
            },
        }
        return cls(config_data)

    @classmethod
    def from_disk(cls, fname: pathlib.Path):
        with open(fname, "r") as f:
            config_data = toml.load(f)
        return cls(config_data)

    def to_disk(self):
        """ Write the config_data object to disk """
        full_cfg_fname = (
            pathlib.Path(__file__).resolve().parents[1]
            / "configs"
            / (self.config_data["cfg_title"] + ".toml")
        )
        with open(full_cfg_fname, "w") as f:
            toml.dump(self.config_data, f)



import pytest
import numpy as np

from pysight import read_lst


class TestReadLst:
    filename = "tests/tests_data/1.lst"
    start_of_data_pos = 1749
    timepatch = "32"
    is_binary = False
    readlst_hex = read_lst.ReadData(
        filename=filename,
        start_of_data_pos=start_of_data_pos,
        timepatch=timepatch,
        is_binary=False,
        debug=False,
    )
    readlst_bin = read_lst.ReadData(
        filename=filename,
        start_of_data_pos=start_of_data_pos,
        timepatch=timepatch,
        is_binary=True,
        debug=False,
    )

    @pytest.mark.parametrize(
        "tp, bytess",
        [
            ("0", 6),
            ("5", 10),
            ("1", 10),
            ("1a", 14),
            ("2a", 14),
            ("22", 14),
            ("32", 14),
            ("2", 14),
            ("5b", 18),
            ("Db", 18),
            ("f3", 18),
            ("43", 18),
            ("c3", 18),
            ("3", 18),
        ],
    )
    def test_get_data_tp_hex(self, tp, bytess):
        assert bytess == self.readlst_hex._get_data_length_bytes(tp, 2)

    @pytest.mark.parametrize(
        "tp, bytess",
        [
            ("0", 2),
            ("5", 4),
            ("1", 4),
            ("1a", 6),
            ("2a", 6),
            ("22", 6),
            ("32", 6),
            ("2", 6),
            ("5b", 8),
            ("Db", 8),
            ("f3", 8),
            ("43", 8),
            ("c3", 8),
            ("3", 8),
        ],
    )
    def test_get_data_tp_bin(self, tp, bytess):
        assert bytess == self.readlst_bin._get_data_length_bytes(tp, 0)

    def test_carriage_hex_windows(self):
        assert 2 == self.readlst_hex._check_carriage_return()

    def test_carriage_bin(self):
        assert 0 == self.readlst_bin._check_carriage_return()

    @pytest.mark.parametrize("debug", [(True,), (False,)])
    def test_det_num_of_lines_hex(self, debug, bytess=14):
        assert int(2.8e6) == self.readlst_hex._determine_num_of_lines(debug, bytess)

    @pytest.mark.parametrize("debug", [(True,), (False,)])
    def test_det_num_of_lines_bin(self, debug, bytess=6):
        assert int(1.2e6) == self.readlst_bin._determine_num_of_lines(debug, bytess)

    def test_ascii_readout(self, data_length=14, num_of_lines=-1):
        data = np.array(
            [
                "0100000060d9\r\n",
                "010000008289\r\n",
                "010000009099\r\n",
                "01000000a3a9\r\n",
                "0100000144f9\r\n",
                "010000018939\r\n",
                "01000001d139\r\n",
                "01000001e8b9\r\n",
            ],
            dtype="<U14",
        )
        assert np.setdiff1d(
            data, self.readlst_hex._read_ascii(data_length, num_of_lines)
        ).shape == (0,)



import attr
from attr.validators import instance_of
import pandas as pd
import sys
import logging


@attr.s(slots=True)
class PhotonDF:
    """
    Create initial photon dataframe and set the channel as its index. It will
    contain a PMT/photons column for every recorded data channel.

    :param Dict[str, pd.DataFrame] dict_of_data: A dictionary with keys
    being the input analog channels and values are the DataFrames that contains the raw data.
    :param bool interleaved: Whether the PMT1 channel is interleaved.
    """

    dict_of_data = attr.ib(validator=instance_of(dict))
    interleaved = attr.ib(validator=instance_of(bool))

    def run(self) -> pd.DataFrame:
        """
        Main function for the class.
        Concatenates all valid data channels into a single dataframe.

        :return pd.DataFrame: Photon data with labeled channels.
        """
        if "Channel" not in self.dict_of_data["PMT1"].columns:
            self.dict_of_data["PMT1"]["Channel"] = 1

        try:
            if "Channel" not in self.dict_of_data["PMT2"].columns:
                self.dict_of_data["PMT2"]["Channel"] = 2
        except KeyError:
            pass

        try:
            df_photons = pd.concat(
                [self.dict_of_data["PMT1"].copy(), self.dict_of_data["PMT2"].copy()],
                axis=0,
            )
        except KeyError:
            df_photons = self.dict_of_data["PMT1"].copy()
        except:
            logging.error("Unknown error: ", sys.exc_info()[0])
        finally:
            df_photons.loc[:, "Channel"] = df_photons.loc[:, "Channel"].astype(
                "category"
            )
            if self.interleaved:
                df_photons["Channel"] = df_photons["Channel"].cat.add_categories(7)
            df_photons.set_index(keys="Channel", inplace=True)

        return df_photons



"""
__author__ = Hagai Hargil
"""

from unittest import TestCase
from pysight.nd_hist_generator.line_signal_validators.validation_tools import (
    SignalValidator,
)
from pysight.nd_hist_generator.line_signal_validators.scanimage import (
    ScanImageLineValidator,
)
import pandas as pd
import numpy as np


class TestScanImage(TestCase):
    """
    Tests for the ScanImageValidator class
    """

    dict_of_data = dict(
        PMT1=pd.DataFrame([1, 10, 20, 30], columns=["abs_time"]),
        Lines=pd.DataFrame([0, 5, 10, 15, 20, 25, 30, 35]),
        columns=["abs_time"],
    )
    vlad = SignalValidator(dict_of_data)

    def test_pairwise(self):
        dict_of_data = {}
        iter = ScanImageLineValidator(
            SignalValidator(dict_of_data)
        )._ScanImageLineValidator__pairwise([1, 2, 3, 4])
        self.assertSequenceEqual([(1, 2), (2, 3), (3, 4)], list(iter))



import numpy as np
import pandas as pd
from pysight.nd_hist_generator.tag_lens import (
    TagPeriodVerifier,
    TagPhaseAllocator,
    TagPipeline,
    numba_digitize,
)


class TestTagPipeline:
    """
    Tests for TAG analysis functions
    """

    tag_data = pd.Series(np.arange(0, 200 * 6530, 6530))
    photons = pd.DataFrame([10, 100, 1000], columns=["abs_time"])
    def_pipe = TagPipeline(photons=photons, tag_pulses=tag_data)

    def test_preservation(self):
        photons = pd.DataFrame([0, 10, 6531], columns=["abs_time"])
        pipe = TagPipeline(photons=photons, tag_pulses=self.tag_data)
        returned = pd.Series([0, 6530])
        assert np.array_equal(
            returned, pipe._TagPipeline__preserve_relevant_tag_pulses()
        )

    def test_preservation_without_zero(self):
        photons = pd.DataFrame([10, 6531], columns=["abs_time"])
        pipe = TagPipeline(photons=photons, tag_pulses=self.tag_data)
        returned = pd.Series([6530])
        assert np.array_equal(
            returned, pipe._TagPipeline__preserve_relevant_tag_pulses()
        )


class TestTagPeriodVerifier:
    """ Test the Verifier class """

    tag_data = pd.Series(np.arange(0, 200 * 6530, 6530))
    freq = 189e3
    binwidth = 800e-12
    def_verifier = TagPeriodVerifier(
        tag=tag_data, freq=freq, binwidth=binwidth, last_photon=np.uint64(200 * 6530)
    )

    def test_bins_bet_pulses(self):
        assert 6614 == self.def_verifier.period

    def test_allowed_noise(self):
        assert 331 == self.def_verifier.allowed_noise

    def test_start_end_no_issues(self):
        tag_data = pd.Series(np.arange(0, 100, 10))
        freq = 0.1
        binwidth = 1.0
        verifier = TagPeriodVerifier(
            tag=tag_data, freq=freq, binwidth=binwidth, last_photon=np.uint64(100)
        )
        ret_start, ret_end = verifier._TagPeriodVerifier__obtain_start_end_idx()
        my_start = np.array([], dtype=np.int64)
        my_end = np.array([], dtype=np.int64)
        assert np.array_equal(my_start, ret_start)
        assert np.array_equal(my_end, ret_end)

    def test_start_end_no_zero(self):
        tag_data = pd.Series(np.arange(0, 300, 10))
        tag_data.drop([0, 5, 6], inplace=True)
        tag_data = tag_data.append(pd.Series([3, 9, 25]))
        tag_data = tag_data.sort_values().reset_index(drop=True)
        freq = 0.1
        binwidth = 1.0
        verifier = TagPeriodVerifier(
            tag=tag_data, freq=freq, binwidth=binwidth, last_photon=np.uint64(300)
        )
        ret_start, ret_end = verifier._TagPeriodVerifier__obtain_start_end_idx()
        my_start = [0, 3, 6]
        my_end = [2, 5, 7]
        assert list(ret_start) == my_start
        assert list(ret_end) == my_end

    def test_start_end_adding_zero(self):
        tag_data = pd.Series(np.arange(5, 300, 10))
        tag_data.drop([1, 5, 7, 8], inplace=True)
        tag_data = tag_data.append(pd.Series([9, 27, 29, 31]))
        tag_data = tag_data.sort_values().reset_index(drop=True)
        freq = 0.1
        binwidth = 1.0
        verifier = TagPeriodVerifier(
            tag=tag_data, freq=freq, binwidth=binwidth, last_photon=np.uint64(300)
        )
        ret_start, ret_end = verifier._TagPeriodVerifier__obtain_start_end_idx()
        my_start = [0, 7]
        my_end = [6, 9]
        assert list(ret_start) == my_start
        assert list(ret_end) == my_end

    def test_fix_tag_pulses_adding_zero(self):
        tag_data = pd.Series(np.arange(0, 100, 10))
        tag_data.drop([0, 5, 6], inplace=True)
        tag_data = tag_data.append(pd.Series([3, 9, 25]))
        tag_data = tag_data.sort_values().reset_index(drop=True)
        freq = 0.1
        binwidth = 1.0
        verifier = TagPeriodVerifier(
            tag=tag_data, freq=freq, binwidth=binwidth, last_photon=np.uint64(100)
        )
        my_start = [0, 3, 6]
        my_end = [2, 5, 7]
        verifier._TagPeriodVerifier__fix_tag_pulses(starts=my_start, ends=my_end)
        assert np.array_equal(verifier.tag.values, np.arange(0, 100, 10))

    def test_fix_tag_pulses_no_zero_end_missing(self):
        tag_data = pd.Series(np.arange(5, 95, 10, dtype=np.uint64))
        tag_data.drop([1, 5, 7, 8], inplace=True)
        tag_data = tag_data.append(pd.Series([9, 27, 29, 31], dtype=np.uint64))
        tag_data = tag_data.sort_values().reset_index(drop=True)
        freq = 0.1
        binwidth = 1.0
        verifier = TagPeriodVerifier(
            tag=tag_data, freq=freq, binwidth=binwidth, last_photon=np.uint64(85)
        )
        my_start = [0, 7]
        my_end = [6, 8]
        verifier._TagPeriodVerifier__fix_tag_pulses(starts=my_start, ends=my_end)
        assert np.array_equal(verifier.tag.values, np.arange(5, 75, 10))


class TestTagPhaseAllocator:
    def test_tag_digitize(self):
        x = np.array([0.2, 6.4, 3.0, 1.6])
        bins = np.array([0.0, 1.0, 2.5, 4.0, 10.0])
        real_result = np.array([1, 4, 3, 2])
        result, _ = numba_digitize(x, bins)
        assert np.array_equal(real_result, result)

    def test_allocate_phase_1(self):
        photons = pd.DataFrame([0, 2 * np.pi, 4 * np.pi], columns=["abs_time"])
        tag = pd.Series([0, 2 * np.pi, 4 * np.pi, 6 * np.pi])
        phaser = TagPhaseAllocator(photons, tag)
        phaser.allocate_phase()
        result = [1, 1, 1]
        for elem1, elem2 in zip(result, phaser.photons.Phase.tolist()):
            assert np.isclose(elem1, elem2)

    def test_allocate_phase_2(self):
        photons = pd.DataFrame([np.pi, 3 * np.pi, 5 * np.pi], columns=["abs_time"])
        tag = pd.Series([0, 2 * np.pi, 4 * np.pi, 6 * np.pi])
        phaser = TagPhaseAllocator(photons, tag)
        phaser.allocate_phase()
        for elem1, elem2 in zip([-1, -1, -1], phaser.photons.Phase.tolist()):
            assert np.isclose(elem1, elem2)

    def test_allocate_phase_3(self):
        photons = pd.DataFrame([1, 2, 3], columns=["abs_time"])
        tag = pd.Series([0, 2 * np.pi, 4 * np.pi, 6 * np.pi])
        phaser = TagPhaseAllocator(photons, tag)
        phaser.allocate_phase()
        normed_result = photons.abs_time / (2 * np.pi)
        true_result = np.sin(normed_result * 2 * np.pi + np.pi / 2).astype(np.float32)
        for elem1, elem2 in zip(true_result.tolist(), phaser.photons.Phase.tolist()):
            assert np.isclose(elem1, elem2, 6)



import unittest
import pandas as pd
import numpy as np
from os import sep
from pysight.ascii_list_file_parser.file_io import ReadMeta
from pysight.ascii_list_file_parser import timepatch_switch


# class TestAllocationTools(unittest.TestCase):
#     length = 100
#     df = pd.DataFrame([i for i in range(length)], columns=['abs_time'])
#     dict_of_data = {
#         'PMT1': pd.DataFrame([[1, 2, 3, 4, 5, 6, 7], [5, 6, 7, 8, 9, 10, 11]],
#                              columns=['abs_time', 'edge', 'sweep', 'time_rel_sweep',
#                                       'time_rel_pulse', 'bins_x', 'bins_y'], dtype=np.uint64),
#         'Lines': pd.DataFrame([[1, 2, 3, 4], [5, 6, 7, 8]],
#                               columns=['abs_time', 'edge', 'sweep', 'time_rel_sweep']),
#         'Laser': pd.DataFrame([[0, 2, 3, 2], [3, 6, 7, 6]],
#                               columns=['abs_time', 'edge', 'sweep', 'time_rel_sweep'])
#     }
#     dict_of_data['PMT1'].time_rel_pulse = dict_of_data['PMT1'].time_rel_pulse.astype(np.uint8)
#     cur_file = ReadMeta(filename='tests' + sep + 'tests_data' + sep + '1.lst')
#     cur_file.run()
#     dict_of_slices_hex = timepatch_switch.ChoiceManagerHex().process(cur_file.timepatch)

# analyzed = Allocate(dict_of_data=cur_file.dict_of_input_channels)
#
# def test_censor_not_needed(self):
#     a = np.array([10, 9, 8, 5])
#     self.assertTrue(not self.analyzed._Analysis__requires_censoring(a))
#
# def test_censor_needed_1(self):
#     a = np.array([10, 9, 0, 5])
#     self.assertTrue(self.analyzed._Analysis__requires_censoring(a))
#
# def test_censor_needed_2(self):
#     a = np.array([10, 9, 0, 7, 5, 0, 2])
#     self.assertTrue(self.analyzed._Analysis__requires_censoring(a))
#
# def test_censor_needed_3(self):
#     a = np.array([10, 9, 0, 7, 5, 0, 2, 5])
#     self.assertTrue(self.analyzed._Analysis__requires_censoring(a))
#
# def test_censor_needed_4(self):
#     a = np.array([10, 9, 0, 7, 5, 0, 2, 0])
#     self.assertTrue(self.analyzed._Analysis__requires_censoring(a))
#
# def test_laser_interpolation(self):
#     pass


if __name__ == "__main__":
    unittest.main()



import json
import logging
from typing import Union, Dict, Any

from pysight.nd_hist_generator.movie import ImagingSoftware


def verify_input(config: Dict[str, Any]):
    """Validate all GUI inputs"""
    data_sources = (
        "PMT1",
        "PMT2",
        "PMT3",
        "PMT4",
        "Lines",
        "Frames",
        "Laser",
        "TAG Lens",
        "Empty",
    )
    channel_inputs = {
        config["inputs"]["start"],
        config["inputs"]["stop1"],
        config["inputs"]["stop2"],
        config["inputs"]["stop3"],
        config["inputs"]["stop4"],
        config["inputs"]["stop5"],
    }
    MINIMAL_TAG_BIT = 1
    MAXIMAL_TAG_BIT = 16

    list_of_keys = [
        config["inputs"]["start"],
        config["inputs"]["stop1"],
        config["inputs"]["stop2"],
        config["inputs"]["stop3"],
        config["inputs"]["stop4"],
        config["inputs"]["stop5"],
    ]
    if "PMT1" not in list_of_keys:
        raise BrokenPipeError("PMT1 value has to be entered in inputs.")

    if config["image"]["num_of_frames"] == None:
        if "Frames" not in data_sources:
            raise BrokenPipeError(
                "You must either enter a frame channel or number of frames."
            )
    else:
        if float(config["image"]["num_of_frames"]) != config["image"]["num_of_frames"]:
            raise ValueError("Please enter an integer number of frames.")

    if config["image"]["num_of_frames"] < 0:
        raise ValueError("Number of frames has to be a positive number.")

    filename = config["outputs"]["data_filename"]
    if not filename.endswith(".lst") and not filename.endswith(".p"):
        raise BrokenPipeError(
            "Please choose a list (*.lst) or pickle (*.p) file for analysis."
        )

    if len(channel_inputs) > len(data_sources):
        raise ValueError(
            "Wrong inputs in channels. Please choose a value from the list."
        )

    set_of_keys = set(list_of_keys)

    if len(list_of_keys) != len(
        set_of_keys
    ):  # making sure only a single option was chosen in the GUI
        non_empty_keys = sorted([x for x in list_of_keys if x != "Empty"])
        sorted_set_of_keys = sorted(list(set_of_keys.difference({"Empty"})))
        if non_empty_keys != sorted_set_of_keys:
            raise KeyError(
                'Input consisted of two or more similar names which are not "Empty".'
            )

    # TAG bits input verification
    tag_bits_group_options = (
        "Power",
        "Slow axis",
        "Fast axis",
        "Z axis",
        "None",
    )
    if config["tagbits"]["tag_bits"]:
        values_of_bits_set = set()
        start_bits_set = set()
        end_bits_set = set()
        for key, val in config["tagbits"].items():
            try:
                groupnum = key[-1]
            except ValueError:
                continue
            cur_label = val["label" + groupnum]
            if cur_label not in tag_bits_group_options:
                raise UserWarning(f"Value {val} not in allowed TAG bits inputs.")
            if not isinstance(val["start" + groupnum], int):
                raise UserWarning(
                    f"The start bit of TAG label {val['start' + groupnum]} wasn't an integer."
                )
            if not isinstance(val["end"] + groupnum, int):
                raise UserWarning(
                    f"The end bit of TAG label {val['end' + groupnum]} wasn't an integer."
                )
            if val["end" + groupnum] < val["start" + groupnum]:
                raise UserWarning(
                    f"Bits in row {groupnum} have a start value which is higher than its end."
                )
            if (
                val["start" + groupnum] > MAXIMAL_TAG_BIT
                or val["end" + groupnum] > MAXIMAL_TAG_BIT
            ):
                raise UserWarning(
                    f"In label {key} maximal allowed TAG bit is {MAXIMAL_TAG_BIT}."
                )
            if (
                val["start" + groupnum] < MINIMAL_TAG_BIT
                or val["end" + groupnum] < MINIMAL_TAG_BIT
            ):
                raise UserWarning(
                    f"In label {key} minimal allowed TAG bit is {MINIMAL_TAG_BIT}."
                )
            values_of_bits_set.add(val["label" + groupnum])
            start_bits_set.add(val["start" + groupnum])
            end_bits_set.add(val["end" + groupnum])

        if len(values_of_bits_set) > len(start_bits_set):
            raise UserWarning("Some TAG bit labels weren't given unique start bits.")

        if len(values_of_bits_set) > len(end_bits_set):
            raise UserWarning("Some TAG bit labels weren't given unique end bits.")

    if not isinstance(config["advanced"]["phase"], float) and not isinstance(
        config["advanced"]["phase"], int
    ):
        raise UserWarning("Mirror phase must be a number.")

    if config["advanced"]["fill_frac"] < 0:
        raise UserWarning("Fill fraction must be a positive number.")

    if not isinstance(config["advanced"]["fill_frac"], float) and not isinstance(
        config["advanced"]["fill_frac"], int
    ):
        raise UserWarning("Fill fraction must be a number.")

    try:
        int(config["image"]["x_pixels"])
        int(config["image"]["x_pixels"])
        int(config["image"]["x_pixels"])
    except ValueError:
        raise UserWarning("Pixels must be an integer number.")

    if config["image"]["x_pixels"] < 0:
        raise UserWarning("X pixels value must be greater than 0.")

    if config["image"]["y_pixels"] < 0:
        raise UserWarning("X pixels value must be greater than 0.")

    if config["image"]["z_pixels"] < 0:
        raise UserWarning("X pixels value must be greater than 0.")

    if float(config["image"]["x_pixels"]) != config["image"]["x_pixels"]:
        raise UserWarning("Enter an integer number for the x-axis pixels.")

    if float(config["image"]["y_pixels"]) != config["image"]["y_pixels"]:
        raise UserWarning("Enter an integer number for the y-axis pixels.")

    if float(config["image"]["z_pixels"]) != config["image"]["z_pixels"]:
        raise UserWarning("Enter an integer number for the z-axis pixels.")

    if config["advanced"]["reprate"] < 0:
        raise UserWarning("Laser repetition rate must be positive.")

    if (config["advanced"]["reprate"] % 10 != 0) and (
        config["advanced"]["flim"] or config["advanced"]["interleaved"]
    ):
        raise UserWarning(
            "FLIM or Deinterleaving isn't supported for laser "
            "reprates that can't be used with a 10 MHz external clock."
            " Please contact package authors."
        )

    if config["advanced"]["binwidth"] < 0:
        raise UserWarning("Binwidth must be a positive number.")

    if config["advanced"]["binwidth"] > 1e-9:
        raise UserWarning("Enter a binwidth with units of [seconds].")

    if type(config["outputs"]["data_filename"]) != str:
        raise UserWarning("Filename must be a string.")

    if "Laser" in channel_inputs and config["advanced"]["flim"] == 1:
        raise UserWarning(
            "Can't have both a laser channel active and the FLIM checkboxed ticked."
        )

    if not config["image"]["imaging_software"].upper() in [
        name for name, member in ImagingSoftware.__members__.items()
    ]:
        raise UserWarning("Must use existing options in the Imaging Software entry.")

    if not isinstance(config["advanced"]["frame_delay"], float):
        raise UserWarning("Frame delay must be a float.")

    if config["advanced"]["frame_delay"] < 0:
        raise UserWarning("Frame delay must be a positive float.")

    if config["advanced"]["frame_delay"] > 10:
        raise UserWarning(
            "Frame delay is the number of seconds between subsequent frames."
        )



from pysight.ascii_list_file_parser.file_io import ReadMeta
import pathlib


class TestMetaTools:
    """ Tests for new multiscaler readout functions """

    list_of_file_names = [
        str(next(pathlib.Path("tests/").rglob("*1.lst")).absolute()),
        str(next(pathlib.Path("tests/").rglob("*2.lst")).absolute()),
    ]
    file_io_objects = []
    for file in list_of_file_names:
        cur_obj = ReadMeta(
            file,
            debug=False,
            input_start="Frames",
            input_stop1="PMT1",
            input_stop2="Lines",
        )
        cur_obj.run()
        file_io_objects.append(cur_obj)

    def test_determine_binary(self):
        list_of_file_binary_formats = [False, False]
        list_of_returned_binaries = []
        for oname in self.file_io_objects:
            list_of_returned_binaries.append(oname.is_binary)

        assert list_of_file_binary_formats == list_of_returned_binaries

    def test_check_range_extraction(self):
        list_of_real_range = [80_000_000, 8064]
        list_of_returned_range = []
        for oname in self.file_io_objects:
            list_of_returned_range.append(oname.data_range)

        assert list_of_real_range == list_of_returned_range

    def test_check_time_patch_extraction(self):
        list_of_real_time_patch = ["32", "5b"]
        list_of_returned_time_patch = []
        for oname in self.file_io_objects:
            metadata = oname._ReadMeta__get_metadata()
            list_of_returned_time_patch.append(oname.timepatch)

        assert list_of_real_time_patch == list_of_returned_time_patch

    def test_check_start_of_data_value(self):
        list_of_real_start_loc = [1749, 1567]
        list_of_returned_locs = []
        for oname in self.file_io_objects:
            list_of_returned_locs.append(oname.start_of_data_pos)

        assert list_of_real_start_loc == list_of_returned_locs

    def test_find_active_channels_from_str(self):
        from test_string_for_fileio import meta

        resdict, res_num_channels = self.file_io_objects[0].find_active_channels(meta)
        assert resdict == {"Frames": "110", "PMT1": "001", "Lines": "010"}
        assert res_num_channels == 1

    def test_find_active_channels_from_bytes(self):
        from test_string_for_fileio import meta_bytes

        self.file_io_objects[0].is_binary = True
        resdict, res_num_channels = self.file_io_objects[0].find_active_channels(
            meta_bytes
        )
        assert resdict == {"Frames": "110", "PMT1": "001", "Lines": "010"}
        assert res_num_channels == 1

    def test_empty_channel_with_data_works(self):
        from test_string_for_fileio import meta

        cur_obj = ReadMeta(
            str(next(pathlib.Path("tests/").rglob("*1.lst")).absolute()),
            debug=False,
            input_start="Empty",
            input_stop1="PMT1",
            input_stop2="Lines",
        )
        cur_obj.is_binary = False
        resdict, res_num_channels = cur_obj.find_active_channels(meta)
        assert resdict == {"PMT1": "001", "Lines": "010"}
        assert res_num_channels == 1

    def test_create_inputs_dict(self):
        real_list_of_real_inputs_dict = [
            {"Frames": "110", "PMT1": "001", "Lines": "010"},
            {"Frames": "110", "PMT1": "001", "Lines": "010"},
        ]
        list_of_returned_inputs_dict = []
        for oname in self.file_io_objects:
            list_of_returned_inputs_dict.append(oname.dict_of_input_channels)

        assert list_of_returned_inputs_dict == real_list_of_real_inputs_dict

    def test_fstchan(self):
        real_fstchan = [0, int((int("54", 16) * 64))]
        returned_fstchan = []
        for oname in self.file_io_objects:
            returned_fstchan.append(oname.acq_delay)

        assert real_fstchan == returned_fstchan

    def test_time_after_sweep(self):
        real_time_after = [960, 960]
        returned_time_after = []
        for oname in self.file_io_objects:
            returned_time_after.append(oname.time_after)

        assert real_time_after == returned_time_after

    def test_bitshifts(self):
        list_of_filenames_for_bitshift = [
            str(next(pathlib.Path("tests/").rglob("*2.lst")).absolute()),
            str(next(pathlib.Path("tests/").rglob("*fly.lst")).absolute()),
            str(next(pathlib.Path("tests/").rglob("*tag.lst")).absolute()),
        ]
        real_bitshifts = [
            2 ** 0,
            2 ** 14,
            2 ** 10,
        ]
        for file, bitshift in zip(list_of_filenames_for_bitshift, real_bitshifts):
            cur_obj = ReadMeta(
                file,
                debug=False,
                input_start="Frames",
                input_stop1="PMT1",
                input_stop2="Lines",
            )
            cur_obj.run()
            assert cur_obj.bitshift == bitshift


if __name__ == "__main__":
    TestMetaTools()



import numpy as np
import pandas as pd

from pysight.nd_hist_generator.deinterleave import Deinterleave


class TestDeinterleave:
    photons = np.array([1, 5, 6, 15, 4, 5, 11, 15, 3, 11, 9], dtype=np.uint8)
    abs_time = np.random.randint(100, 1000, photons.shape, dtype=np.uint64)
    df = (
        pd.DataFrame(
            {
                "time_rel_pulse": photons,
                "abs_time": abs_time,
                "Channel": np.ones_like(photons),
                "Lines": np.ones_like(photons),
                "Frames": np.zeros_like(photons),
            }
        )
        .assign(Channel=lambda x: x.Channel.astype("category"))
        .assign(Channel=lambda x: x.Channel.cat.add_categories(7))
        .set_index(["Channel", "Lines", "Frames"], drop=True)
    )

    deinter = Deinterleave(df, reprate=0.25, binwidth=0.25)
    deinter_photons = deinter.run()

    def test_no_photon_left_behind(self):
        assert len(self.deinter_photons) == len(self.photons)

    def test_correct_allocation_early(self):
        idx_slice = pd.IndexSlice
        early_photons = self.deinter_photons.loc[idx_slice[1, :, :], :].copy()
        assert np.array_equal(
            early_photons["time_rel_pulse"].values.ravel(),
            np.array([1, 5, 6, 4, 5, 3], dtype=np.uint8),
        )

    def test_correct_allocation_late(self):
        idx_slice = pd.IndexSlice
        late_photons = self.deinter_photons.loc[idx_slice[7, :, :], :].copy()
        assert np.array_equal(
            late_photons["time_rel_pulse"].values.ravel(),
            np.array([15, 11, 15, 11, 9], dtype=np.uint8),
        )

    def test_abs_time_unchanged(self):
        assert np.array_equal(
            self.deinter_photons["abs_time"].values.ravel().sort(), self.abs_time.sort()
        )



from typing import Dict, Any, Tuple, Union
from collections import namedtuple
import pathlib
import logging

import attr
from attr.validators import instance_of
import numpy as np
import pandas as pd
import zarr
from numcodecs import Blosc


def trunc_end_of_file(name) -> str:
    """
    Take only the start of the filename to avoid error with Python and Windows

    :param str name: Filename to truncate
    :return str:
    """
    return name[:240]


@attr.s(slots=True)
class OutputParser:
    """
    Parse the wanted outputs and produce a dictionary with
    file pointers to the needed outputs
    # TODO: Handle the interleaved case
    """

    output_dict = attr.ib(validator=instance_of(dict))
    filename = attr.ib(validator=instance_of(str), converter=trunc_end_of_file)
    num_of_frames = attr.ib(default=1, validator=instance_of(int))
    x_pixels = attr.ib(default=512, validator=instance_of(int))
    y_pixels = attr.ib(default=512, validator=instance_of(int))
    z_pixels = attr.ib(default=1, validator=instance_of(int))
    channels = attr.ib(
        default=pd.CategoricalIndex([1]), validator=instance_of(pd.CategoricalIndex)
    )
    binwidth = attr.ib(default=100e-12, validator=instance_of(float))
    reprate = attr.ib(default=80e6, validator=instance_of(float))
    lst_metadata = attr.ib(factory=dict, validator=instance_of(dict))
    file_pointer_created = attr.ib(default=True, validator=instance_of(bool))
    cache_size = attr.ib(default=10 * 1024 ** 3, validator=instance_of(int))
    debug = attr.ib(default=False, validator=instance_of(bool))
    #: Dictionary with data - either the full or summed stack, with a list of zarr objects as channels
    outputs = attr.ib(init=False)
    #: Tuple of (num_of_frames, x, y, z, tau)
    data_shape = attr.ib(init=False)
    num_of_channels = attr.ib(init=False)

    def run(self):
        """ Parse what the user required, creating a list of zarr dataset pointers for each channel """
        self.outputs = {}
        self.num_of_channels = len(self.channels)
        self.data_shape = self.determine_data_shape_full()
        if not self.output_dict:
            return
        if self.output_dict["memory"]:
            self.outputs["memory"] = 1
        self.__create_prelim_file()
        if self.file_pointer_created:
            self.__populate_file()

    @property
    def _group_names(self):
        return {
            "summed": ["Summed Stack/Channel 1", "Summed Stack/Channel 2"],
            "stack": ["Full Stack/Channel 1", "Full Stack/Channel 2"],
            "flim": ["Lifetime/Channel 1", "Lifetime/Channel 2"],
        }

    def __create_prelim_file(self):
        """ Try to create a preliminary zarr file. Cache improves IO performance """
        if (
            self.output_dict["stack"]
            or self.output_dict["summed"]
            or self.output_dict["flim"]
        ):
            try:
                path = pathlib.Path(self.filename)
                newpath = path.with_suffix(".zarr")
                debugged = "_DEBUG" if self.debug else ""
                fullfile = newpath.with_name(newpath.stem + debugged + newpath.suffix)
                self.outputs["filename"] = zarr.open(str(fullfile), mode="w")
                self.outputs["filename"].attr = self.lst_metadata
            except (PermissionError, OSError):
                self.file_pointer_created = False
                logging.warning("Permission Error: Couldn't write data to disk.")
            else:
                self.file_pointer_created = True
                self.outputs.update(
                    {key: True for key, val in self.output_dict.items() if val}
                )

    def _create_compressor(self):
        """Generate a compressor object for the Zarr array"""
        return Blosc(cname="zstd", clevel=3, shuffle=Blosc.BITSHUFFLE)

    def __populate_file(self):
        """
        Generate files and add metadata to each group, write out the data in chunks
        f: zarr file pointer
        """
        data_shape_summed = self.data_shape[1:]
        chunk_shape = list(self.data_shape)
        chunk_shape[0] = 1
        chunk_shape = tuple(chunk_shape)
        if self.output_dict["stack"]:
            try:
                self.__create_group(
                    output_type="stack",
                    shape=self.data_shape,
                    chunks=chunk_shape,
                    dtype=np.uint8,
                )
            except (PermissionError, OSError):
                self.file_pointer_created = False
        if self.output_dict["summed"]:
            try:
                self.__create_group(
                    output_type="summed",
                    shape=data_shape_summed,
                    chunks=True,
                    dtype=np.uint16,
                )
            except (PermissionError, OSError):
                self.file_pointer_created = False

        if self.output_dict["flim"]:
            try:
                self.__create_group(
                    output_type="flim",
                    shape=self.data_shape,
                    chunks=chunk_shape,
                    dtype=np.float32,
                )
            except (PermissionError, OSError):
                self.file_pointer_created = False
        if self.file_pointer_created is False:
            logging.warning("Permission Error: Couldn't write data to disk.")

    def __create_group(
        self,
        output_type: str,
        shape: tuple,
        chunks: Union[Tuple, bool],
        dtype: np.dtype,
    ):
        """Create a group in the open file with the given parameters."""
        groupname = self._group_names[output_type]
        for channel in range(self.num_of_channels):
            self.outputs[output_type] = self.outputs["filename"].zeros(
                groupname[channel],
                shape=shape,
                dtype=dtype,
                chunks=chunks,
                compressor=self._create_compressor(),
            )

    def determine_data_shape_full(self):
        """
        Return the tuple that describes the shape of the final dataset.
        Dimension order: [FRAME, X, Y, Z, LIFETIME]
        """
        non_squeezed = (
            self.x_pixels,
            self.y_pixels,
            self.z_pixels,
        )
        squeezed_shape = tuple([dim for dim in non_squeezed if dim != 1])
        return (
            self.num_of_frames,
        ) + squeezed_shape  # we never "squeeze" the number of frames


DataShape = namedtuple("DataShape", "t, x, y, z")


@attr.s(frozen=True)
class PySightOutput:
    """
    Keeps the relevant data from the run of the algorithm for later
    in-memory processing.

    :param pd.DataFrame photons: The 'raw' photon DataFrame.
    :param dict _summed_mem: Summed-over-time arrays of the data - one per channel.
    :param dict _stack: Full data arrays, one per channel.
    :param pd.CategoricalIndex _channels: Actual data channels analyzed.
    :param tuple _data_shape: Data dimensions
    :param bool _flim: Whether data has Tau channel.
    :param Dict[str,Any] config: Configuration file used in this run.
    """

    # TODO: Add new FLIM handler

    photons = attr.ib(validator=instance_of(pd.DataFrame), repr=False)
    _summed_mem = attr.ib(validator=instance_of(dict), repr=False)
    _stack = attr.ib(validator=instance_of(dict), repr=False)
    _channels = attr.ib(validator=instance_of(pd.CategoricalIndex), repr=False)
    _data_shape = attr.ib(validator=instance_of(tuple), repr=False)
    _flim = attr.ib(validator=instance_of(bool), repr=False)
    config = attr.ib(validator=instance_of(dict), repr=False)
    available_channels = attr.ib(init=False)
    data_shape = attr.ib(init=False)
    ch1 = attr.ib(init=False, repr=False)
    ch2 = attr.ib(init=False, repr=False)
    ch3 = attr.ib(init=False, repr=False)
    ch4 = attr.ib(init=False, repr=False)
    ch5 = attr.ib(init=False, repr=False)
    ch6 = attr.ib(init=False, repr=False)
    ch7 = attr.ib(init=False, repr=False)

    def __attrs_post_init__(self):
        """
        Populate the different attributes of the class
        """
        object.__setattr__(self, "available_channels", list(self._channels))
        object.__setattr__(self, "data_shape", self._parse_data_shape())
        for channel in self._channels:
            cur_stack = MultiDimensionalData(
                self._stack[channel], self._summed_mem[channel], self.data_shape
            )
            object.__setattr__(self, "ch" + str(channel), cur_stack)

    def _parse_data_shape(self):
        """
        Turns the data shape tuple into a namedtuple
        """
        shape = self._data_shape[:3]
        if len(self._data_shape) == 4:  # take TAG shape regardless
            shape += (self._data_shape[3],)
        else:
            shape += (None,)

        return DataShape(*shape)


@attr.s(frozen=True)
class MultiDimensionalData:
    """
    Internal representation of a stack of data.

    :param np.ndarray full: The entirety of the data.
    :param np.ndarray time_summed: All data summed across the \
        time dimension.
    :param DataShape _data_shape: List of valid dimensions.
    :param np.ndarray z_summed: All data summed across the \
        z dimension.
    """

    full = attr.ib(validator=instance_of(np.ndarray), repr=False)
    time_summed = attr.ib(validator=instance_of(np.ndarray), repr=False)
    _data_shape = attr.ib(validator=instance_of(DataShape))
    z_summed = attr.ib(init=False, repr=False)

    def __attrs_post_init__(self):
        if self._data_shape.z:
            object.__setattr__(self, "z_summed", self.full.sum(axis=3))



import attr
from attr.validators import instance_of
import numpy as np
import logging


@attr.s
class ReadData:
    """
    Read a given .lst file into memory

    :param str filename: File to read
    :param int start_of_data_pos: Number of header bytes in the file to skip
    :param str timepatch: Type of file
    :param bool is_binary: Whether the file is a binary or ASCII file
    :param bool debug: Run a debug build (limited number of lines for quick execution of the entire pipeline)
    """

    filename = attr.ib(validator=instance_of(str))
    start_of_data_pos = attr.ib(validator=instance_of(int))
    timepatch = attr.ib(validator=instance_of(str))
    is_binary = attr.ib(validator=instance_of(bool))
    debug = attr.ib(validator=instance_of(bool))
    data = attr.ib(init=False)

    def __attrs_post_init__(self):
        assert self.filename != ""
        assert self.start_of_data_pos > 0
        assert self.timepatch != ""

    @property
    def data_length(self):
        """  Number of bits in a word for each timepatch """
        return {
            "0": 16,
            "5": 32,
            "1": 32,
            "1a": 48,
            "2a": 48,
            "22": 48,
            "32": 48,
            "2": 48,
            "5b": 64,
            "Db": 64,
            "f3": 64,
            "43": 64,
            "c3": 64,
            "3": 64,
        }

    def read_lst(self) -> np.ndarray:
        """
        Main method to load the .lst file into memory.

        :return np.ndarray: An array with the ascii or binary data \
        that was read to memory using one of the helper functions.
        """
        bytes_to_add = self._check_carriage_return()
        data_length_bytes = self._get_data_length_bytes(self.timepatch, bytes_to_add)
        num_of_lines_in_bits = self._determine_num_of_lines(
            self.debug, data_length_bytes
        )
        # Read file
        if self.is_binary:
            self.data = self._read_binary(data_length_bytes, num_of_lines_in_bits)
        else:
            self.data = self._read_ascii(data_length_bytes, num_of_lines_in_bits)

        logging.info("File read. Sorting the file according to timepatch...")
        return self.data

    def _check_carriage_return(self) -> int:
        r""" If the file was generated in Windows each ASCII word will end with
        '\r\n' (2 bytes). Else it will probably end with '\n'.  """
        if self.is_binary:
            return 0
        with open(self.filename, "rb") as f:
            f.seek(self.start_of_data_pos)
            arr = np.fromfile(f, dtype="18S", count=18).astype("18U")
        first = arr[0]
        if "\r" in first:
            return 2
        else:
            return 1

    def _get_data_length_bytes(self, tp: str, bytes_to_add: int):
        r"""
        Number of bytes per word for each timepatch. For the hex
        case it takes into consideration the fact that each two letters
        correspond to a single byte.
        In addition, Windows-generated lst files end with '\r\n', while
        Mac/Linux ones might only have '\n', so we make sure to add these
        bytes in as well.
        """
        if self.is_binary:
            return self.data_length[tp] // 8
        else:
            return self.data_length[tp] // 4 + bytes_to_add  # 2 is \r\n, 1 is just \n

    def _determine_num_of_lines(self, debug, bytess):
        """ In debug runs we don't read all lines """
        if debug:
            num_of_lines = int(0.2e6 * bytess)  # 200k events is usually enough
            printstr = f'[DEBUG] Reading file "{self.filename}"...'
        else:
            num_of_lines = -1
            printstr = f'Reading file "{self.filename}"...'
        logging.info(printstr)
        return num_of_lines

    def _read_binary(self, data_length, num_of_lines) -> np.ndarray:
        with open(self.filename, "rb") as f:
            f.seek(self.start_of_data_pos)
            arr: np.ndarray = np.fromfile(
                f, dtype=f"u{data_length}", count=num_of_lines
            )
        return arr

    def _read_ascii(self, data_length, num_of_lines) -> np.ndarray:
        with open(self.filename, "rb") as f:
            f.seek(self.start_of_data_pos)
            arr = np.fromfile(
                f, dtype="{}S".format(data_length), count=num_of_lines
            ).astype("{}U".format(data_length))
        return arr



from typing import List, Dict, Tuple
import functools
import operator

import attr
from attr.validators import instance_of
import numpy as np
import pandas as pd
from scipy.optimize import curve_fit
from scipy.signal import find_peaks

from pysight.nd_hist_generator.outputs import DataShape


@attr.s(slots=True)
class FrameChunk:
    """
    Holds a chunk of data and can histogram it efficiently.
    Composed out of a Movie - any attribute or method not defined here is
    taken from the self.movie object with the __getattr__ method
    """

    movie = attr.ib()
    df_dict = attr.ib(validator=instance_of(dict), repr=False)
    frames = attr.ib(validator=instance_of(pd.Series))
    lines = attr.ib(validator=instance_of(pd.Series))
    frames_per_chunk = attr.ib(validator=instance_of(int))
    hist_dict = attr.ib(init=False)
    end_time = attr.ib(init=False)

    def __getattr__(self, item):
        return getattr(self.movie, item)

    def __attrs_post_init__(self):
        if len(self.frames) > 1:
            dif = np.uint64(self.frames.diff().mean())
            self.end_time = self.frames.iloc[-1] + dif
        else:
            self.end_time = self.frames.iloc[0] + 1

    def create_hist(self) -> Dict[int, Tuple[np.ndarray]]:
        """
        Main method to create the histogram of data. Assigns each event
        in the dataframe to its correct location, for each channel.

        :return: A dictionary with its keys being the spectral channels
                of the data, and the value is a tuple of the histogrammed data and the edges.
        """
        self.hist_dict: Dict[int, Tuple[np.ndarray]] = {}
        for chan in self.df_dict:
            list_of_edges = self.__create_hist_edges(chan)
            data_columns: List[np.ndarray] = []
            data_columns.append(self.df_dict[chan]["abs_time"].to_numpy())
            data_columns.append(self.df_dict[chan]["time_rel_line"].to_numpy())
            try:
                data_columns.append(self.df_dict[chan]["Phase"].to_numpy())
            except KeyError:
                pass
            if self.flim:
                hist, flim_hist = self._hist_with_flim(
                    data_columns, list_of_edges, chan
                )
            else:
                flim_hist = None
                hist, _ = np.histogramdd(sample=data_columns, bins=list_of_edges)
            hists = self._post_process_hist([hist.astype(np.uint8)])
            # TODO: Throw this away once we do FLIM properly
            all_hists = hists + (flim_hist,)
            self.hist_dict[chan] = all_hists
        return self.hist_dict

    def _post_process_hist(self, hists: List[np.ndarray]):
        processed = []
        for hist in hists:
            if hist is not None:
                reshaped = hist.reshape(
                    (self.frames_per_chunk,) + (self.data_shape[1:])
                )
                if self.bidir:
                    reshaped[:, 1::2, ...] = np.flip(reshaped[:, 1::2, ...], axis=2)
                processed.append(reshaped)
            else:
                processed.append(None)
        return tuple(processed)

    def __create_hist_edges(self, chan) -> List[np.ndarray]:
        """
        Generate the grid of the histogram.

        :param int chan: Channel number

        :return ``list`` of ``np.ndarray``: One for each dimension
        """
        edges = []
        edges.append(self.__create_line_edges())
        edges.append(self.__create_col_edges())

        if "Phase" in self.df_dict[chan].columns:
            edges.append(self.__linspace_along_sine())

        return edges

    def __censor_correction(self, hist):
        raise NotImplementedError(
            "No censor correction as of yet. Contact package authors."
        )

    def __create_frame_edges(self) -> np.ndarray:
        """ Create edges for a numpy histogram for the frames dimension """

        assert self.frames.shape[0] == self.frames_per_chunk
        frames = np.hstack((self.frames.values, self.frames.values[-1] + np.uint64(1)))
        return frames

    def __create_line_edges(self) -> np.ndarray:
        """ Takes existing lines and turns them into bin edges. """

        assert (
            self.lines.shape[0] <= self.x_pixels * self.frames_per_chunk
        )  # last chunk can have less frames
        all_lines = np.hstack(
            (self.lines.values, self.lines.values[-1] + self.line_delta)
        )
        return all_lines

    def __create_col_edges(self) -> np.ndarray:
        if self.x_pixels == 1:
            return np.linspace(
                0, self.end_time, num=self.y_pixels + 1, endpoint=True, dtype=np.uint64
            )

        delta = self.line_delta if self.bidir else self.line_delta / 2
        start = 0
        if self.image_soft == "ScanImage":
            start = delta * 0.04
        col_end = delta * self.fill_frac / 100 if self.fill_frac > 0 else delta

        return np.linspace(
            start=start,
            stop=int(col_end),
            num=self.y_pixels + 1,
            endpoint=True,
            dtype=np.uint64,
        )

    def __linspace_along_sine(self) -> np.ndarray:
        """
        Find the points that are evenly spaced along a sine function between pi/2 and 3*pi/2

        :return: Array of bin edges
        """
        lower_bound = -1 if self.tag_as_phase else 0
        upper_bound = 1 if self.tag_as_phase else self.tag_period
        pts = []
        relevant_idx = []

        bin_edges = np.linspace(
            lower_bound, upper_bound, self.z_pixels + 1, endpoint=True
        )[:, np.newaxis]
        dx = 0.00001
        x = np.arange(np.pi / 2, 3 * np.pi / 2 + dx, step=dx, dtype=np.float64)
        sinx = np.sin(x)
        locs = np.where(np.isclose(sinx, bin_edges, atol=1e-05))
        vals, first_idx, count = np.unique(
            locs[0], return_index=True, return_counts=True
        )
        assert len(vals) == len(bin_edges)
        for first_idx, count in zip(first_idx, count):
            idx_to_append = locs[1][first_idx + count // 2]
            relevant_idx.append(idx_to_append)
            pts.append(sinx[idx_to_append])

        return np.array(pts)

    def _hist_with_flim(
        self, data: List[np.ndarray], edges: List[np.ndarray], chan: int
    ) -> Tuple[np.ndarray, Tuple[np.ndarray]]:
        """Run a slightly more complex processing pipeline when we need to calculate
        the lifetime of each pixel in the image.
        We first generate the standard histogram without taking the FLIM dimension
        into consideration, but we do keep the bin index of each of the photons
        in the image.
        Then we groupby the photons based on their hist index, and each such group
        goes to function which calculates the decay curve constants there.

        Parameters
        ----------
        data : list of np.ndarray
            Photon arrival times in each of the dimensions
        edges : list of np.ndarray
            Histogram edges for each dimension.
        chan : int
            Channel number

        Returns
        -------
        hist : np.ndarray
            N-dimensional histogram, where N = len(data)
        hist_with_flim : np.ndarray
            N-dimensional histogram, where N = len(data)
        """
        hist = HistWithIndex(data, edges)
        hist.run()
        valid_photons = hist.discard_out_of_bounds_photons()
        # TODO
        # flim = FlimCalc(self.df_dict[chan]["time_rel_pulse"].to_numpy(), hist.hist_indices)
        # flim.run()
        # only_flim_hist = np.full_like(hist.hist_photons, np.nan)
        # only_flim_hist[flim.hist_arrivals["bin"]] = flim.hist_arrivals["lifetime"]
        # assert only_flim_hist.shape == hist.hist_photons.shape
        return (
            hist.hist_photons,
            pd.DataFrame(
                {
                    "since_laser": self.df_dict[chan]["time_rel_pulse"].to_numpy()[
                        valid_photons
                    ],
                    "bin": hist.hist_indices[valid_photons],
                }
            ),
        )


@attr.s
class HistWithIndex:
    """A 'manual' implementation of np.histogramdd which also returns
    the indices of the partitioned photons, so that we could take these
    indices and copy them to be used with other data - the lifetime
    of these pixels, in our case.

    Parameters
    --------
    data : list of np.ndarray
        The data in a list, each dimension represnted as a vector of arrival times

    edges : list of np.ndarray
        The edges into which we should bin the photons
    """

    data = attr.ib(validator=instance_of(list))
    edges = attr.ib(validator=instance_of(list))
    hist_photons = attr.ib(init=False)
    hist_indices = attr.ib(init=False)
    edges_with_outliers = attr.ib(init=False)

    def __attrs_post_init__(self):
        """A few asserts regarding the data's shape and validity."""
        if len(self.data) != len(self.edges):
            raise TypeError(
                f"Data and bin length are unequal. Data length is {len(self.data)} while edges are {len(self.edges)}."
            )

    def run(self):
        """Main class pipeline."""
        self.hist_indices, self.edges_with_outliers = self._get_indices_for_photons()
        self.hist_photons = self._populate_hist_with_photons(self.edges_with_outliers)

    def _get_indices_for_photons(self):
        """For the given data and edges, find the indices in which each photon should
        belong. This is the first step in computing a histogram, and can be thought
        of as 'arghistogramdd'.
        """
        indices = tuple(
            np.searchsorted(edges, self.data[idx], side="right")
            for idx, edges in enumerate(self.edges)
        )

        nedges = []
        for dim, edge in enumerate(self.edges):
            on_edge = self.data[dim] == edge[-1]
            indices[dim][on_edge] -= 1
            nedges.append(len(edge) + 1)

        nedges = np.array(nedges)
        return np.ravel_multi_index(indices, nedges), nedges

    def _populate_hist_with_photons(self, nedges: np.ndarray):
        """Populates an empty n-d array with the photons in the indices
        which were calculate in "_get_indices_for_photons".

        Parameters
        --------

        nedges : np.ndarray
            number of edges per dimension

        """
        hist = np.bincount(self.hist_indices, minlength=nedges.prod())
        hist = hist.reshape(nedges)

        # remove outliers
        core = len(self.edges) * (slice(1, -1),)
        hist = hist[core]
        return hist

    def discard_out_of_bounds_photons(self):
        """The "run" methods automatically discards these photons, but if
        we wish to create a histogram by hand we have to discard them
        manually as well.
        This method has to be run after self._get_indices_for_photons has.
        """
        unraveled = np.unravel_index(self.hist_indices, self.edges_with_outliers)
        outliers = []
        for indices, edge in zip(unraveled, self.edges_with_outliers):
            outliers.append((indices != 0) & (indices < edge - 1))
        valid_photons = outliers[0]
        for indices in range(1, len(outliers)):
            valid_photons = np.logical_and(valid_photons, outliers[indices])
        return valid_photons


@attr.s(hash=True)
class FlimCalc:
    """An object designed to calculate the lifetime decay constant
    of the generated movie.
    It receives as input the photon arrival times and their binning index, and
    it then bins these arrival times and calculates the parameters of the
    observed exponential decay curve which rises from these bins. The final
    value of each of the bins is the tau calculated from that fit.

    Parameters
    --------
    data : np.ndarray
        The arrival times of all photons in the experiment

    indices : np.ndarray
        The bin indices of each of the photons

    downsample : int, optional
        How much downsampling should be conducted on the stack.
    """

    data = attr.ib(validator=instance_of(np.ndarray))
    indices = attr.ib(validator=instance_of(np.ndarray))
    downsample = attr.ib(default=10, validator=instance_of(int))
    bins_bet_pulses = attr.ib(default=125, validator=instance_of(int))
    all_data = attr.ib(init=False)

    def __attrs_post_init__(self):
        self.all_data = pd.DataFrame({"since_laser": self.data, "bin": self.indices})

    def run(self):
        """Run the calculation pipeline."""
        self._partition_photons_into_bins()
        self._normalize_taus_to_uint8()

    def _partition_photons_into_bins(self):
        """Once we have the indices where each photon belongs, we can cluster them and
        calculate the lifetime of them all. This method clusters the photons into
        groups and sends this group off for lifetime calculation. The partitioning
        is dependent on the downsampling factor required by the user.
        """
        self.hist_arrivals = self.all_data.groupby(
            "bin", as_index=False, sort=False
        ).agg(calc_lifetime)

    def _normalize_taus(self):
        """FLIM images will be displayed in a float32 scale
        due to the nans.
        """
        self.hist_arrivals["lifetime"] = (self.hist_arrivals["since_laser"]).astype(
            np.float32
        )

    def histogram_result(self, shape: DataShape):
        """Create a histogram with the bins and lifetimes for each of the photons
        that were calculated in the "run" pipeline.

        Parameters
        ----------
        shape : DataShape
            Shape of data

        Returns
        -------
        hist : np.ndarray
            The histogrammed data
        """
        total_bins = functools.reduce(operator.mul, shape, 1)
        assert total_bins >= self.hist_arrivals["bin"]
        hist = np.full(shape, np.nan, dtype=np.float32).ravel()
        hist[self.hist_arrivals["bin"]] = self.hist_arrivals["lifetime"]
        hist = hist.reshape(shape)
        core = (len(shape) - 1) * (slice(1, -1),)
        core = (slice(None),) + core
        hist = hist[core]
        return hist


def calc_lifetime(data: pd.Series, bins_bet_pulses=124) -> float:
    """Calculate the lifetime of the given data by fitting it to a decaying exponent
    with a lifetime around 3 ns.
    # TODO: bins_bet_pulses
    """
    if len(data) < 5:
        return np.nan
    hist = np.histogram(data, bins_bet_pulses)[0]
    if hist.max() < 5:
        return np.nan
    peaks, props = find_peaks(hist, height=(None, None), prominence=(0.8, None))
    # If no peak is found we'll try using the first bin as a peak.
    # scipy.signal.find_peaks is not good at detecting that the first
    # data point is the highest. If it's not true then the curve fit
    # will eventually fail, leaving us with a nan instead of tau.
    if len(peaks) == 0:
        peaks, props = np.array([0]), {"peak_heights": hist[0]}
    decay_curve, max_val, min_val = find_decay_borders(hist, peaks, props)
    if len(decay_curve) < 4:
        return np.nan
    try:
        popt, _ = curve_fit(
            _exp_decay,
            np.arange(len(decay_curve)),
            decay_curve,
            p0=(max_val, 1 / 35, min_val),
            maxfev=1_000,
        )
    except RuntimeError:
        return np.nan
    tau = np.array(1 / popt[1]).astype(np.float32, casting="safe")
    if (tau > bins_bet_pulses) or (tau < 0):
        return np.nan
    return tau


def _exp_decay(x, a, b, c):
    """ Exponential function for FLIM and censor correction """
    return a * np.exp(-b * x) + c


def find_decay_borders(
    hist: np.ndarray, peaks: np.ndarray, props: Dict[str, np.ndarray]
):
    """Trims a given histogram which contains an exponential decay curve so that
    it starts at the peak and ends at the lowest point after it."""
    highest_peak_idx = peaks[np.argmax(props["peak_heights"])]
    decay_curve = hist[highest_peak_idx:]
    end_of_decay = np.argmin(decay_curve)
    decay_curve = decay_curve[: end_of_decay + 1]
    return decay_curve, decay_curve[0], decay_curve[-1]



from enum import Enum
from typing import NamedTuple, Dict, Any
import logging
import numpy as np
import pandas as pd
import attr
from attr.validators import instance_of

from pysight.ascii_list_file_parser.file_io import ReadMeta


class TimepatchBits(NamedTuple):
    total: int
    time: int
    sweep: int
    tag: int
    lost: int


class Timepatch(Enum):
    Tp0 = TimepatchBits(16, 12, 0, 0, 0)
    Tp5 = TimepatchBits(32, 20, 8, 0, 0)
    Tp1 = TimepatchBits(32, 28, 0, 0, 0)
    Tp5b = TimepatchBits(64, 28, 16, 15, 1)
    TpDb = TimepatchBits(64, 28, 16, 16, 0)
    Tpf3 = TimepatchBits(64, 36, 7, 16, 1)
    Tp43 = TimepatchBits(64, 44, 0, 15, 1)
    Tpc3 = TimepatchBits(64, 44, 0, 16, 0)
    Tp3 = TimepatchBits(64, 54, 0, 5, 1)


@attr.s
class BinaryDataParser:
    """ Interpret data from binary list files """

    data = attr.ib(validator=instance_of(np.ndarray))
    timepatch = attr.ib(validator=instance_of(str))
    data_range = attr.ib(default=0, validator=instance_of(int))
    bitshift = attr.ib(default=0, validator=instance_of(int))
    acq_delay = attr.ib(default=0, validator=instance_of(int))
    holdafter = attr.ib(default=0, validator=instance_of(int))
    dict_of_inputs = attr.ib(default=attr.Factory(dict), validator=instance_of(dict))
    use_tag_bits = attr.ib(default=False, validator=instance_of(bool))
    dict_of_inputs_bin = attr.ib(init=False)
    data_to_grab = attr.ib(init=False)
    channel = attr.ib(init=False)
    edge = attr.ib(init=False)
    time = attr.ib(init=False)
    tag = attr.ib(init=False)
    sweep = attr.ib(init=False)
    lost = attr.ib(init=False)
    dict_of_data = attr.ib(init=False)
    timepatch_bits = attr.ib(init=False)
    aligned_data = attr.ib(init=False)

    def __attrs_post_init__(self):
        self.data_to_grab = ["abs_time"]
        tpdict = {
            "0": Timepatch.Tp0,
            "5": Timepatch.Tp5,
            "1": Timepatch.Tp1,
            "5b": Timepatch.Tp5b,
            "Db": Timepatch.TpDb,
            "f3": Timepatch.Tpf3,
            "43": Timepatch.Tp43,
            "c3": Timepatch.Tpc3,
            "3": Timepatch.Tp3,
        }
        try:
            self.timepatch_bits = tpdict[self.timepatch]
        except KeyError:
            raise KeyError(f"Invalid timepatch value received: {self.timepatch}.")

        self.dict_of_inputs_bin = {}
        for key, val in self.dict_of_inputs.items():
            self.dict_of_inputs_bin[key] = int(val, 2)

    def run(self):
        """ Main pipeline for the parsing """
        self.channel = self.__get_channel()
        self.__check_user_inputs()
        self.edge = self.__get_edge()
        self.time = self.__get_time()
        if self.timepatch_bits.value.sweep != 0:
            self.sweep = self.__get_sweep()
            # self.sweep = self.__unfold_sweeps(self.sweep)
        if self.timepatch != "f3":
            if self.timepatch_bits.value.tag != 0:
                self.tag = self.__get_tag()
            if self.timepatch_bits.value.lost != 0:
                self.lost = self.__get_lost()
        else:
            self.tag = self.__get_tag_f3()
            self.lost = self.__get_lost_f3()
        self.aligned_data = self.__gen_df()
        self.dict_of_data = self.__slice_df_to_dict()
        logging.info(
            "Sorted dataframe created. Starting to set the proper data channel distribution..."
        )

    def __get_channel(self) -> np.ndarray:
        """
        Parse the channel and edge information from the data.

        :return np.ndarray chan: Array of channel numbers, with 6 == START
        """
        chan = (self.data & 0b111).astype(np.uint8)
        if np.any(chan > 6):
            logging.warning(
                f"Illegal channels found in file. Encountered the following"
                f" values: {np.unique(chan)}.\nTrying to continue."
            )
        return chan

    def __get_edge(self) -> np.ndarray:
        """
        Parse the edge of each data line.

        :return np.ndarray edge: Array of edges, 1 means falling edge
        """
        return (np.right_shift(self.data, 3) & 1).astype(np.uint8)

    def __get_time(self) -> np.ndarray:
        """
        Parse the time bits.

        :return np.ndarray time: Array of the absolute times for each event.
        """
        ones = int("1" * self.timepatch_bits.value.time, 2)  # time bits
        time = np.right_shift(self.data, 4) & ones
        return time.astype(np.uint64)

    def __get_sweep(self) -> np.ndarray:
        """
        Parse the sweep bits.

        :return np.ndarray sweep: Array of the sweep number for each event.
        """
        ones = int("1" * self.timepatch_bits.value.sweep, 2)
        right_shift_by = 4 + self.timepatch_bits.value.time
        sweep = np.right_shift(self.data, right_shift_by) & ones
        return sweep.astype(np.uint16)

    def __unfold_sweeps(self, sweeps: np.ndarray) -> np.ndarray:
        """In some cases more than the real sweep number in the list file
        exceeds the available numbers of represented sweeps. That is, an
        experiment can be coducted using 1000 sweeps, while the sweep counter
        only has 8 bits. This causes 'folding' in the sweep values. The goal of
        this method is to unfold the sweep numbers such that they're (almost)
        monotonically increasing.
        """
        max_possible_sweep_value = (2 ** self.timepatch_bits.value.sweep) - 1
        if sweeps.max() < max_possible_sweep_value:
            return sweeps
        folding_idx = np.where(sweeps == max_possible_sweep_value)
        if len(folding_idx) == 1:
            return sweeps

    def __get_tag(self) -> np.ndarray:
        """
        Parse the tag bits.

        :return np.ndarray tag: Array of the tag bits for each event.
        """
        ones = int("1" * self.timepatch_bits.value.tag, 2)
        right_shift_by = (
            4 + self.timepatch_bits.value.time + self.timepatch_bits.value.sweep
        )
        tag = np.right_shift(self.data, right_shift_by) & ones
        return tag.astype(np.uint16)

    def __get_lost(self) -> np.ndarray:
        """
        Parse the lost bits.

        :return np.ndarray lost: Array of the lost bit for each event.
        """
        right_shift_by = self.timepatch_bits.value.total - 1
        lost = np.right_shift(self.data, right_shift_by) & 1
        return lost.astype(np.uint8)

    def __get_tag_f3(self) -> np.ndarray:
        """
        Parse the TAG bits of the f3 timepatch files.

        :return np.ndarray tag: Array of the TAG bits for each event.
        """
        tag = np.right_shift(self.data, 48) & 65535
        return tag.astype(np.uint16)

    def __get_lost_f3(self) -> np.ndarray:
        """
        Parse the lost bit of the f3 timepatch files.

        :return np.ndarray lost: Array of the lost bit for each event.
        """
        lost = np.right_shift(self.data, 47) & 1
        return lost.astype(np.uint8)

    def __check_user_inputs(self):
        """
        Assert that the channels that the user believe were recorded are actually there.
        Before sorting all photons make sure that no input is missing from the user. If it's missing
        the code will ignore this channel, but not raise an exception
        """

        actual_data_channels = set(np.unique(self.channel)).difference({0})
        if actual_data_channels != set(self.dict_of_inputs_bin.values()):
            logging.warning(
                "Channels that were inserted in GUI don't match actual data channels recorded. \n"
                f"The list files contains data in the following channels: {actual_data_channels}."
            )
            thrown_channels = 0
            keys_to_pop = []
            for key, item in self.dict_of_inputs_bin.items():
                if item not in actual_data_channels:
                    keys_to_pop.append(key)
                    thrown_channels += 1
            [self.dict_of_inputs_bin.pop(key) for key in keys_to_pop]

    def __gen_df(self) -> pd.DataFrame:
        """
        Align the acquired data into a single DataFrame

        :return pd.DataFrame:
        """
        df = pd.DataFrame(
            self.time, index=[self.channel, self.edge], columns=["abs_time"]
        )
        try:
            df.abs_time += (self.sweep - 1) * (
                (self.data_range * self.bitshift) + self.holdafter
            ) + (self.sweep * self.acq_delay)
        except AttributeError:
            pass
        try:
            tag_ser = pd.Series(self.tag, index=[self.channel, self.edge])
            df["tag"] = tag_ser
        except AttributeError:
            pass
        else:
            if self.use_tag_bits:
                self.data_to_grab.extend(["tag", "edge"])

        try:
            lost_ser = pd.Series(self.lost, index=[self.channel, self.edge])
            df["lost"] = lost_ser
        except AttributeError:
            pass

        df.index.names = ["analog_input", "edge"]
        return df

    def __slice_df_to_dict(self) -> dict:
        """
        Take the DataFrame of data and create a dictionary of data from it
        :return dict:
        """
        dict_of_data = {}
        for key, analog_chan in self.dict_of_inputs_bin.items():
            relevant_vals = self.aligned_data.xs(key=analog_chan, level=0).loc[
                :, self.data_to_grab
            ]
            if key in ["PMT1", "PMT2"]:
                dict_of_data[key] = relevant_vals.reset_index(drop=True)
                dict_of_data[key]["Channel"] = (
                    1 if "PMT1" == key else 2
                )  # channel is the spectral channel
            else:
                dict_of_data[key] = relevant_vals.sort_values(
                    by=["abs_time"]
                ).reset_index(drop=True)

        return dict_of_data


def binary_parsing(cur_file: ReadMeta, raw_data: np.ndarray, config: Dict[str, Any]):
    """Reads a binary file to memory."""
    binary_parser = BinaryDataParser(
        data=raw_data,
        data_range=cur_file.data_range,
        timepatch=cur_file.timepatch,
        bitshift=cur_file.bitshift,
        acq_delay=cur_file.acq_delay,
        holdafter=cur_file.time_after,
        use_tag_bits=config["tagbits"]["tag_bits"],
        dict_of_inputs=cur_file.dict_of_input_channels,
    )
    binary_parser.run()
    return binary_parser.data_to_grab, binary_parser.dict_of_data



"""
#!/usr/bin/env python
# -*- coding: utf-8 -*-
__author__ = Hagai Har-Gil
"""

from typing import Dict, Any, Optional
import pickle
import logging
import pathlib
import sys

logging.basicConfig(
    stream=sys.stdout,
    # filename=str(pathlib.Path('.') / "general.log"),
    # filemode="w",
    format="%(levelname)s :: %(filename)s :: %(asctime)s :: %(message)s",
    level=logging.INFO,
)

import matplotlib
import pandas as pd
import colorama
import numpy as np

matplotlib.rcParams["backend"] = "TkAgg"
import matplotlib.pyplot as plt
import toml

colorama.init()
from pysight.ascii_list_file_parser.file_io import ReadMeta, ascii_parsing
from pysight.nd_hist_generator.allocation import Allocate
from pysight.nd_hist_generator.outputs import OutputParser, PySightOutput
from pysight.nd_hist_generator.movie import Movie
from pysight.nd_hist_generator.gating import GatedDetection
from pysight.nd_hist_generator.photon_df import PhotonDF
from pysight.nd_hist_generator.tag_bits import ParseTAGBits
from pysight.nd_hist_generator.line_signal_validators.validation_tools import (
    SignalValidator,
)
from pysight.nd_hist_generator.line_signal_validators.add_bidir_lines import (
    add_bidir_lines,
)
from pysight.gui.gui_helpers import verify_input
from pysight.gui.config_parser import Config
from pysight.nd_hist_generator.volume_gen import VolumeGenerator
from pysight.binary_list_file_parser.binary_parser import (
    BinaryDataParser,
    binary_parsing,
)
from pysight.read_lst import ReadData
from pysight.nd_hist_generator.deinterleave import Deinterleave


def main_data_readout(config: Dict[str, Any]) -> Optional[PySightOutput]:
    """
    Main function that reads the lst file and processes its data.
    Should not be run independently - only from other "run_X" functions.

    :param dict config: Loaded configuration file as a dictionary.

    :return PySightOutput: An object containing the relevant data,\
    if "memory" option was checked in the GUI.
    """
    if config["outputs"]["data_filename"].endswith(".lst"):
        relevant_columns, dict_of_data, lst_metadata, fill_frac = _read_lst_file(config)
    else:
        logging.info(f"Reading file {config['outputs']['data_filename']}...")
        with open(config["outputs"]["data_filename"], "rb") as f:
            dict_of_data = pickle.load(f)
        lst_metadata = dict()
        relevant_columns = ["abs_time"]
        fill_frac = config["advanced"]["fill_frac"]

    validated_data = SignalValidator(
        dict_of_data=dict_of_data,
        num_of_frames=config["image"]["num_of_frames"],
        binwidth=float(config["advanced"]["binwidth"]),
        use_sweeps=config["advanced"]["sweeps_as_lines"],
        delay_between_frames=float(config["advanced"]["frame_delay"]),
        data_to_grab=relevant_columns,
        line_freq=config["advanced"]["line_freq"],
        num_of_lines=config["image"]["x_pixels"],
        bidir=config["advanced"]["bidir"],
        bidir_phase=config["advanced"]["phase"],
        image_soft=config["image"]["imaging_software"],
        lst_metadata=lst_metadata,
    )

    validated_data.run()

    photon_df = PhotonDF(
        dict_of_data=validated_data.dict_of_data,
        interleaved=config["advanced"]["interleaved"],
    )
    photons = photon_df.run()
    tag_bit_parser = ParseTAGBits(
        dict_of_data=validated_data.dict_of_data,
        photons=photons,
        use_tag_bits=config["tagbits"]["tag_bits"],
        bits_dict=config["tagbits"],
    )

    if not config["advanced"]["bidir"]:
        validated_data.dict_of_data = add_bidir_lines(validated_data.dict_of_data)
    analyzed_struct = Allocate(
        bidir=config["advanced"]["bidir"],
        tag_offset=config["advanced"]["tag_offset"],
        laser_freq=float(config["advanced"]["reprate"]),
        binwidth=float(config["advanced"]["binwidth"]),
        tag_pulses=int(config["advanced"]["tag_pulses"]),
        phase=config["advanced"]["phase"],
        keep_unidir=config["advanced"]["keep_unidir"],
        flim=config["advanced"]["flim"],
        censor=config["advanced"]["censor"],
        dict_of_data=validated_data.dict_of_data,
        df_photons=tag_bit_parser.gen_df(),
        tag_freq=float(config["advanced"]["tag_freq"]),
        tag_to_phase=True,
        deinterleave=config["advanced"]["interleaved"],
    )
    analyzed_struct.run()
    data_for_movie = analyzed_struct.df_photons

    del photons
    del photon_df
    if config["advanced"]["interleaved"]:
        logging.warning(
            """Deinterleaving a data channel is currently highly experimental and
            is supported only on data in the PMT1 channel. Inexperienced users
            are highly advised not to use it."""
        )
        deinter = Deinterleave(
            photons=analyzed_struct.df_photons,
            reprate=config["advanced"]["reprate"],
            binwidth=config["advanced"]["binwidth"],
        )
        data_for_movie = deinter.run()
    # Determine type and shape of wanted outputs, and open the file pointers there
    outputs = OutputParser(
        num_of_frames=len(validated_data.dict_of_data["Frames"]),
        output_dict=config["outputs"],
        filename=config["outputs"]["data_filename"],
        x_pixels=config["image"]["x_pixels"],
        y_pixels=config["image"]["y_pixels"],
        z_pixels=config["image"]["z_pixels"] if analyzed_struct.tag_interp_ok else 1,
        channels=data_for_movie.index.levels[0],
        binwidth=config["advanced"]["binwidth"],
        reprate=config["advanced"]["reprate"],
        lst_metadata=lst_metadata,
        debug=config["advanced"]["debug"],
    )
    outputs.run()

    line_delta = validated_data.line_delta
    del validated_data
    if config["advanced"]["gating"]:
        logging.warning(
            "Gating is currently not implemented. Please contact package authors."
        )
        # gated = GatedDetection(
        #     raw=analyzed_struct.df_photons, reprate=config['advanced']['reprate'], binwidth=config['advanced']['binwidth']
        # )
        # gated.run()
    # Create a movie object
    volume_chunks = VolumeGenerator(
        frames=analyzed_struct.dict_of_data["Frames"], data_shape=outputs.data_shape
    )
    frame_slices = volume_chunks.create_frame_slices()
    final_movie = Movie(
        data=data_for_movie,
        frames=analyzed_struct.dict_of_data["Frames"],
        frame_slices=frame_slices,
        num_of_frame_chunks=volume_chunks.num_of_chunks,
        reprate=float(config["advanced"]["reprate"]),
        name=config["outputs"]["data_filename"],
        data_shape=outputs.data_shape,
        binwidth=float(config["advanced"]["binwidth"]),
        bidir=config["advanced"]["bidir"],
        fill_frac=fill_frac,
        outputs=outputs.outputs,
        censor=config["advanced"]["censor"],
        mirror_phase=config["advanced"]["phase"],
        lines=analyzed_struct.dict_of_data["Lines"],
        channels=data_for_movie.index.levels[0],
        flim=config["advanced"]["flim"] or config["advanced"]["interleaved"],
        lst_metadata=lst_metadata,
        line_delta=int(line_delta),
        tag_as_phase=True,
        tag_freq=float(config["advanced"]["tag_freq"]),
        image_soft=config["image"]["imaging_software"],
        frames_per_chunk=volume_chunks.frames_per_chunk,
    )

    final_movie.run()
    if "memory" in outputs.outputs:
        pysight_output = PySightOutput(
            photons=data_for_movie,
            summed_mem=final_movie.summed_mem,
            stack=final_movie.stack,
            channels=data_for_movie.index.levels[0],
            data_shape=outputs.data_shape,
            flim=config["advanced"]["flim"],
            config=config,
        )
        return pysight_output


def _read_lst_file(config: Dict[str, Any]):
    """Read out the data in the .lst file. This function won't be called
    if PySight was called on already-parsed data, i.e. data which came
    from a source other than a multiscaler.
    """
    cur_file = ReadMeta(
        filename=config["outputs"]["data_filename"],
        input_start=config["inputs"]["start"],
        input_stop1=config["inputs"]["stop1"],
        input_stop2=config["inputs"]["stop2"],
        input_stop3=config["inputs"]["stop3"],
        input_stop4=config["inputs"]["stop4"],
        input_stop5=config["inputs"]["stop5"],
        binwidth=config["advanced"]["binwidth"],
        use_sweeps=config["advanced"]["sweeps_as_lines"],
        mirror_phase=config["advanced"]["phase"],
    )
    cur_file.run()
    raw_data_obj = ReadData(
        filename=config["outputs"]["data_filename"],
        start_of_data_pos=cur_file.start_of_data_pos,
        timepatch=cur_file.timepatch,
        is_binary=cur_file.is_binary,
        debug=config["advanced"]["debug"],
    )
    raw_data = raw_data_obj.read_lst()
    if cur_file.is_binary:
        relevant_columns, dict_of_data = binary_parsing(cur_file, raw_data, config)
    else:
        relevant_columns, dict_of_data = ascii_parsing(cur_file, raw_data, config)
    lst_metadata = cur_file.lst_metadata
    fill_frac = (
        config["advanced"]["fill_frac"]
        if cur_file.fill_fraction == -1.0
        else cur_file.fill_fraction
    )
    return relevant_columns, dict_of_data, lst_metadata, fill_frac


def mp_main_data_readout(config: Dict[str, Any]):
    """
    Wrapper for main_data_readout that
    wraps it with a try block. To be used with the
    multiprocessing run option.
    """
    try:
        out = main_data_readout(config)
    except Exception:
        pass
    else:
        return out


def run(cfg_file: str = None) -> Optional[PySightOutput]:
    """ Run PySight.

    :param str cfg_file: Optionally supply an existing configuration filename. Otherwise a GUI will open.

    :return PySightOutput: Object containing raw and processed data
    """
    from pysight.gui.gui_main import GuiAppLst

    if cfg_file:
        with open(cfg_file, "r") as f:
            config: Dict[str, Any] = toml.load(f)
    else:
        gui = GuiAppLst()
        gui.root.mainloop()
        config = Config.from_gui(gui).config_data
    verify_input(config)
    return main_data_readout(config)


def run_batch_lst(
    foldername: str,
    glob_str: str = "*.lst",
    recursive: bool = False,
    cfg_file: str = "",
) -> pd.DataFrame:
    """
    Run PySight on all list files in the folder

    :param str foldername: - Main folder to run the analysis on.
    :param str glob_str: String for the `glob` function to filter list files
    :param bool recursive: Whether the search should be recursive.
    :param str cfg_file: Name of config file to use
    :return pd.DataFrame: Record of analyzed data
    """
    from pysight.gui.gui_main import GuiAppLst

    path = pathlib.Path(foldername)
    num_of_files = 0
    if not path.exists():
        raise UserWarning(f"Folder {foldername} doesn't exist.")
    if recursive:
        all_lst_files = path.rglob(glob_str)
        logging.info(f"Running PySight on the following files:")
        for file in list(all_lst_files):
            logging.info(str(file))
            num_of_files += 1
        all_lst_files = path.rglob(glob_str)
    else:
        all_lst_files = path.glob(glob_str)
        logging.info(f"Running PySight on the following files:")
        for file in list(all_lst_files):
            logging.info(str(file))
            num_of_files += 1
        all_lst_files = path.glob(glob_str)

    data_columns = ["fname", "done", "error"]
    data_record = pd.DataFrame(
        np.zeros((num_of_files, 3)), columns=data_columns
    )  # store result of PySight
    try:
        with open(cfg_file, "r") as f:
            config = toml.load(f)
            config["outputs"]["data_filename"] = ".lst"
    except (TypeError, FileNotFoundError):
        gui = GuiAppLst()
        gui.root.mainloop()
        gui.filename.set(".lst")  # no need to choose a list file
        config = Config.from_gui(gui).config_data
    verify_input(config)

    try:
        for idx, lst_file in enumerate(all_lst_files):
            config["outputs"]["data_filename"] = str(lst_file)
            data_record.loc[idx, "fname"] = str(lst_file)
            try:
                main_data_readout(config)
            except BaseException as e:
                logging.warning(
                    f"File {str(lst_file)} returned an error. Moving onwards."
                )
                data_record.loc[idx, "done"] = False
                data_record.loc[idx, "error"] = repr(e)
            else:
                data_record.loc[idx, "done"] = True
                data_record.loc[idx, "error"] = None
    except TypeError as e:
        logging.error(repr(e))

    logging.info(f"Summary of batch processing:\n{data_record}")
    return data_record


def mp_batch(
    foldername, glob_str="*.lst", recursive=False, n_proc=None, cfg_file: str = ""
):
    """
    Run several instances of PySight using the multiprocessing module.

    :param str foldername: Folder to scan
    :param str glob_str: Glob string to filter files
    :param bool recursive: Whether to scan subdirectories as well
    :param int n_proc: Number of processes to use (None means all)
    :param str cfg_file: Configuration file name
    """
    import pathlib
    import multiprocessing as mp

    from pysight.gui.gui_main import GuiAppLst

    path = pathlib.Path(foldername)
    if not path.exists():
        raise UserWarning(f"Folder {foldername} doesn't exist.")
    if recursive:
        all_lst_files = path.rglob(glob_str)
    else:
        all_lst_files = path.glob(glob_str)
    logging.info(f"Running PySight on the following files:")
    for file in list(all_lst_files):
        logging.info(str(file))

    all_lst_files = path.rglob(glob_str) if recursive else path.glob(glob_str)
    try:
        with open(cfg_file, "r") as f:
            config = toml.load(f)
            config["outputs"]["data_filename"] = ".lst"
    except (TypeError, FileNotFoundError):
        gui = GuiAppLst()
        gui.root.mainloop()
        gui.filename.set(".lst")  # no need to choose a list file
        config = Config.from_gui(gui).config_data
    verify_input(config)
    all_cfgs = []
    for file in all_lst_files:
        config["outputs"]["data_filename"] = str(file)
        all_cfgs.append(config)
    with mp.Pool(n_proc) as pool:
        pool.map(mp_main_data_readout, all_cfgs)


if __name__ == "__main__":
    out = run()
